from tko.util.text import Text
from tko.util.raw_terminal import RawTerminal
from tko.util.runner import Runner
from tko.play.images import compilling_image
from tko.run.solver_builder import SolverBuilder
import subprocess
import random
import sys
import os

if os.name == 'nt':  # Windows
    import msvcrt

    def input_available():
        return msvcrt.kbhit()

    def read_input():
        return msvcrt.getwch()  # use getch() para byte
else:  # Unix (Linux, macOS)
    import select

    def input_available():
        return select.select([sys.stdin], [], [], 0)[0]

    def read_input():
        return sys.stdin.readline()

class Free:
    @staticmethod
    def free_run(solver: SolverBuilder, show_compilation:bool=True, to_clear: bool=True, wait_input:bool=True, header: Text = Text()) -> bool:

        if to_clear:
            Runner.clear_screen()
        if show_compilation:
            image = random.choice(list(compilling_image.keys()))
            for line in compilling_image[image].splitlines():
                print(Text().addf("y", line).center(RawTerminal.get_terminal_size(), Text.Token(" ")))

        if show_compilation:
            Runner.clear_screen()
        solver.prepare_exec()
        if solver.has_compile_error():
            executable, _ = solver.get_executable()
            print(executable.get_error_msg())
        else:
            executable, _ = solver.get_executable()
            cmd, folder = executable.get_command()
            if header.len() == 0:
                print(Text().center(RawTerminal.get_terminal_size(), Text.Token("─")))
            else:
                print(header.center(RawTerminal.get_terminal_size(), Text.Token("─")))
            answer = subprocess.run(cmd, cwd=folder, shell=True if isinstance(cmd, str) else False, text=True)
            if answer.returncode != 0 and answer.returncode != 1:
                print(Runner.decode_code(answer.returncode))
        solver.reset()
        to_run_again = False
        if wait_input:
            while input_available():
                read_input()
            print(Text().center(RawTerminal.get_terminal_size(), Text.Token("─")))
            print(Text.format("Para [recompilar e] reexecutar pressione {y}", "enter"))
            print(Text.format("Para voltar para tela anterior digite {y} e pressione {y}", "q", "enter"))
            # clear buffer before get input

            valor = input()
            if valor != "n" and valor != "q":
                if to_clear:
                    Runner.clear_screen()
                to_run_again = True
        if to_clear:
            Runner.clear_screen()

        return to_run_again



from tko.logger.log_item_base import LogItemBase
from tko.logger.log_sort import LogSort
from tko.logger.delta import Delta

class TaskListener:
    def __init__(self):
        self.minutes_limit: int = 60
        self.time_item: list[tuple[Delta, LogItemBase]] = []
        self.task_dict: dict[str, LogSort] = {}
        self.last_key: str = ""

    def handle_log_entry(self, item: LogItemBase, new_entry: bool = False):
        _ = new_entry
        without_inc_time = Delta.Mode(Delta.Mode.Action.without_inc_time)
        LogSort.add_to_list(without_inc_time, self.time_item, item)
        self.add_to_task(item)

    def add_to_task(self, item: LogItemBase):
        cumulative = True
        delta, _ = self.time_item[-1]
        if delta.elapsed.total_seconds() / 60 > self.minutes_limit:
            cumulative = False
        
        new_key = item.get_key()
        old_key = self.last_key
        if new_key != "" and old_key != "" and new_key != old_key:
            cumulative = False
                  
        if new_key != "":
            self.last_key = new_key

        if new_key not in self.task_dict:
            self.task_dict[new_key] = LogSort()
        
        if cumulative:
            mode = Delta.Mode(Delta.Mode.Action.incrementing_time)
        else:
            mode = Delta.Mode(Delta.Mode.Action.without_inc_time)
        self.task_dict[new_key].add_item(mode, item)



from tko.settings.rep_paths import RepPaths
from tko.settings.rep_source import RepSource
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.util.text import Text
from pathlib import Path

class RepSourceActions:
    def __init__(self, folder: Path | None):
        # if folder is set, use folder, else use remote if remote, else .
        if folder is not None:
            folder = Path(folder).resolve()
        else:
            folder = Path.cwd()

        path = RepPaths.rec_search_for_repo_parents(folder)
        if path is None:
            raise ValueError(f"Repositório não encontrado em {folder}")
        self.folder = folder
        self.repo = Repository(path).load_config()
    
    def remote_ls(self):
        rep = self.repo
        sources = rep.data.get_sources()
        print("Você também pode configurar as fontes e filtros manualmente editando o arquivo:")
        print(Text.format("{y}", rep.paths.get_config_file()))
        if len(sources) == 0:
            print("Nenhuma fonte configurada")
            return
        print("Fontes configuradas:")
        for source in sources:
            print(Text.format("- Rótulo: {y}", source.name))
            print(Text.format("  - Link ou Caminho: {y}", source.get_url_link()))
            print(Text.format("  - File Path     : {y}", source.get_source_readme()))
            print(Text.format("  - Filtragem      : {y}", "Desativado" if source.quests is None else 'Ativado'))
            for f in source.quests or []:
                print(f"    - {f}")

    def remote_rm(self, alias: str) -> None:
        rep = self.repo
        sources = rep.data.get_sources()
        found = False
        for source in sources:
            if source.name == alias:
                found = True
                rep.data.del_source(alias)
                print(Text.format("Fonte {y} removida com sucesso.", alias))
                break
        if not found:
            raise Warning("fail: fonte não encontrada.")
        rep.save_config()

    def remote_filter(self, alias: str, quests: list[str] | None = None, tasks: list[str] | None = None, clear: bool = False) -> None:
        rep = self.repo
        sources = rep.data.get_sources()
        found = False
        for source in sources:
            if source.name == alias:
                found = True
                if quests is not None:
                    if source.quests is None:
                        source.quests = []
                    for f in quests:
                        if f not in source.quests:
                            source.quests.append(f)
                if tasks is not None:
                    if source.tasks is None:
                        source.tasks = []
                    for f in tasks:
                        if f not in source.tasks:
                            source.tasks.append(f)
                if clear:
                    source.quests = None
                    source.tasks = None
                print(Text.format("Filtros {y} atualizados com sucesso.", alias))
                break
        if not found:
            raise Warning("fail: fonte não encontrada.")
        rep.save_config()

    def remote_add(self, name: str, remote_default: str | None, remote_url: str | None, remote_dir: str | None, quest_filter: list[str] | None, task_filter: list[str] | None, writeable: bool, branch: str = "master") -> None:
        
        if any(source.name == name for source in self.repo.data.get_sources()):
            raise Warning("fail: fonte com esse nome já existe.")

        repo = self.repo
        if remote_default is not None:
            print(Text.format("Adicionando fonte remota apontando para repositório git remoto {y}.", remote_default))
            url: str = ""
            settings = Settings()
            if not settings.has_alias_git(remote_default):
                raise Warning("fail: alias git remoto não encontrado.")
            url = settings.get_alias_git(remote_default)
            self.git_clone_repository(url)
            self.repo.data.set_source(RepSource(alias=name).set_git_source(target=url, branch=branch).set_filters(quest_filter, task_filter).set_writeable(writeable))
        elif remote_dir is not None:
            dir_path = Path(remote_dir)
            if not dir_path.exists() or not dir_path.is_dir():
                raise Warning("fail: diretório remoto não encontrado.")
            print(Text.format("Adicionando fonte remota apontando parao repositório no diretorio {y}.", dir_path))
            repo.data.set_source(RepSource(alias=name).set_local_source(target=dir_path).set_filters(quests=quest_filter, tasks=task_filter).set_writeable(writeable))
        elif remote_url is not None:
            print(Text.format("Adicionando fonte remota apontando para repositório git remoto {y}.", remote_url))
            self.git_clone_repository(remote_url)
            self.repo.data.set_source(RepSource(alias=name).set_git_source(target=remote_url, branch=branch).set_filters(quest_filter, task_filter).set_writeable(writeable))

        self.repo.save_config()
   

    def git_clone_repository(self, link: str) -> None:
        print(Text.format("Clonando repositório remoto {y}.", link))
        _ = self.repo.cache.get(link, force_update=True)
        print(Text.format("Repositório {y} clonado com sucesso.", link))
        


    def print_end_msg(self):
        print(Text.format("Voce pode acessar o repositório com o comando {g}", "tko open"))
        



from tko.logger.log_item_base import LogItemBase
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_move import LogItemMove
from tko.logger.log_item_self import LogItemSelf
from tko.logger.delta import Delta
from icecream import ic # type: ignore

class LogSort:
    def __init__(self):
        self.base_list: list[tuple[Delta, LogItemBase]] = []
        self.exec_list: list[tuple[Delta, LogItemExec]] = []
        self.diff_list: list[tuple[Delta, LogItemExec]] = []
        self.move_list: list[tuple[Delta, LogItemMove]] = []
        self.self_list: list[tuple[Delta, LogItemSelf]] = []

    def add_item(self, mode: Delta.Mode, item: LogItemBase) -> None:
        delta = LogSort.add_to_list(mode, self.base_list, item)
        self.__sort_by_instance(delta, item)
    
    def __sort_by_instance(self, delta: Delta, base: LogItemBase) -> None:
        if isinstance(base, LogItemExec):
            self.exec_list.append((delta, base))
            if base.get_size() > 0:
                self.diff_list.append((delta, base))
        elif isinstance(base, LogItemMove):
            self.move_list.append((delta, base))
        elif isinstance(base, LogItemSelf):
            self.self_list.append((delta, base))
        else:
            raise ValueError("Invalid Item Type")
        

    @staticmethod
    def add_to_list(
        mode: Delta.Mode, base_list: list[tuple[Delta, LogItemBase]], item: LogItemBase
    ) -> Delta:
        last_delta: Delta | None = None
        if base_list:
            last_delta, _ = base_list[-1]
        delta = Delta().create(mode, last_delta, item.get_datetime())
        base_list.append((delta, item))
        return delta


import csv

class CSV:
    quest_char = '='
    @staticmethod
    def format_sheet(sheet: list[list[str]]) -> list[list[str]]:
        nl = len(sheet)
        nc = len(sheet[0])

        for c in range(nc):
            for l in range(nl):
                sheet[l][c] = sheet[l][c].strip()

        for c in range(nc):
            max_len = 0
            for l in range(nl):
                max_len = max(max_len, len(sheet[l][c]))
            for l in range(nl):
                fillchar = ' ' if sheet[l][c][0] != CSV.quest_char else CSV.quest_char
                if c == 0:
                    if fillchar == CSV.quest_char:
                        sheet[l][c] = sheet[l][c].ljust(max_len, fillchar)
                    else:
                        sheet[l][c] = sheet[l][c].ljust(max_len, fillchar)
                else:
                    sheet[l][c] = sheet[l][c].rjust(max_len, fillchar)
        return sheet
    
    @staticmethod
    def load_csv(arquivo_csv: str) -> list[list[str]]:
        try:
            with open(arquivo_csv, "r", newline="", encoding="utf-8") as csvfile:
                reader = csv.reader(csvfile)
                sheet = list(reader)
                return sheet
        except FileNotFoundError:
            exit(f"Arquivo '{arquivo_csv}' não encontrado.")

    @staticmethod
    def save_csv(arquivo_csv: str, sheet: list[list[str]]):
        try:
            with open(arquivo_csv, "w", newline="", encoding="utf-8") as csvfile:
                writer = csv.writer(csvfile)
                writer.writerows(sheet)
        except FileNotFoundError:
            exit(f"Arquivo '{arquivo_csv}' não encontrado.")


import os

from tko.enums.execution_result import ExecutionResult
from tko.run.diff_builder_down import DiffBuilderDown
from tko.run.wdir import Wdir
from tko.enums.diff_count import DiffCount
from tko.util.param import Param
from tko.run.diff_builder_side import DiffBuilderSide
from tko.util.text import Text

from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols

from tko.util.freerun import Free
from tko.play.tester import Tester
from tko.play.tracker import Tracker
from tko.run.unit_runner import UnitRunner
from tko.game.task import Task
from tko.play.opener import Opener
from tko.settings.settings import Settings
from tko.enums.diff_mode import DiffMode
from tko.feno.filter import CodeFilter
from tko.settings.repository import Repository
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_sort import LogSort
from tko.logger.task_resume import TaskResume
from tko.settings.rep_paths import RepPaths
from icecream import ic # type: ignore
from pathlib import Path

class TkoFilterMode:
    @staticmethod
    def deep_copy_and_change_dir():
        # path to ~/.tko_filter
        filter_path = os.path.join(os.path.expanduser("~"), ".tko_filter")

        CodeFilter.cf_recursive(".", filter_path, force=True)
        os.chdir(filter_path)

class Run:
    EVAL_MODE_PAD = 14

    def __init__(self, settings: Settings, target_list: list[Path | str], param: None | Param.Basic):
        self.settings = settings
        self.target_list: list[Path] = [Path(target) for target in target_list]
        if param is None:
            self.param = Param.Basic()
        else:
            self.param = param
        self.wdir: Wdir = Wdir(self.settings)
        self.wdir_builded = False
        self.__curses_mode: bool = False
        self.__lang = ""
        self.__rep: Repository | None = None
        self.__task: Task | None = None
        self.__track_folder: Path | None = None
        self.__opener: Opener | None = None
        self.__run_without_ask: bool = True
        self.__show_track_info: bool = False
        self.__show_self_info: bool = False
        self.__eval_mode: bool = False
        self.__complex_percent: bool = False
        self.__abord_on_exec_error: bool = False
        self.__no_run: bool = False
        self.__timeout: int = 0

    def set_eval_mode(self):
        self.__eval_mode = True
        return self

    def show_track_info(self):
        self.__show_track_info = True
        return self
    
    def set_no_run(self):
        self.__no_run = True
        return self
    
    def show_self_info(self):
        self.__show_self_info = True
        return self

    def set_complex_percent(self):
        self.__complex_percent = True
        return self

    def set_abort_on_exec_error(self):
        self.__abord_on_exec_error = True
        return self
    
    def set_timeout(self, timeout: int):
        self.__timeout = timeout
        return self

    def set_curses(self, value:bool=True):
        self.__curses_mode = value
        return self
   
    def set_lang(self, lang:str):
        self.__lang = lang
        return self
    
    def set_opener(self, opener: Opener):
        self.__opener = opener
        return self

    def set_run_without_ask(self, value:bool):
        self.__run_without_ask = value
        return self

    def set_task(self, rep: Repository, task: Task):
        self.__rep = rep
        self.__task = task
        return self
    
    def get_task(self) -> Task:
        if self.__task is None:
            raise Warning("Task não definida")
        return self.__task

    def execute(self):
        if len(self.target_list) == 0:
            if self.__rep is None:
                self.__try_load_rep(Path())
            self.__try_load_task(Path())

        elif len(self.target_list) == 1 and os.path.isdir(self.target_list[0]):
            if self.__rep is None:
                self.__try_load_rep(Path(self.target_list[0]))
            self.__try_load_task(Path(self.target_list[0]))

        if not self.wdir_builded:
            self.build_wdir()

        if self.__missing_target():
            return 0

        if not self.wdir.has_solver() and self.wdir.has_tests() and not self.__eval_mode:
            self.__list_mode()
            return 0
        if self.wdir.has_solver():
            self.__prepare_rep_task_logger()

        if self.wdir.has_solver() and not self.wdir.has_tests() and not self.__curses_mode and not self.__no_run:
            if not self.__eval_mode:
                self.__free_run()
            else:
                print(Text().addf("", "fail: ") + "Nenhum caso de teste encontrado.")
            return 0
        return self.__run_tests()

    def __prepare_rep_task_logger(self):
        if self.__rep is None:
            solver_path = self.wdir.get_solver().args_list[0]
            dirname = Path(os.path.dirname(os.path.abspath(solver_path)))
            self.__try_load_rep(dirname)
            self.__try_load_task(dirname)

        if self.__task is None:
            self.__fill_task()

    def __try_load_rep(self, dirname: Path) -> bool:
        repo_path = RepPaths.rec_search_for_repo_parents(dirname)
        if repo_path is None:
            return False
        rep: Repository = Repository(repo_path)
        rep.load_config()
        rep.load_game()
        self.__rep = rep
        if rep.data.get_lang() != "":
            self.__lang = rep.data.get_lang()
        return True

    def __try_load_task(self, dirname: Path) -> bool:
        if self.__rep is None:
            return False
        rep: Repository = self.__rep
        task_key = rep.get_key_from_task_folder(dirname)
        if task_key == "":
            return False
        task = rep.game.tasks.get(task_key)
        if task is None:
            return False
        self.__task = task
        self.__track_folder = rep.paths.get_track_task_folder(task_key)
        return True

    def __remove_duplicates(self):
        # remove duplicates in target list keeping the order
        self.target_list = list(dict.fromkeys(self.target_list))

    def __change_targets_to_filter_mode(self) -> None:
        if not self.param.filter:
            return

        old_dir: Path = Path.cwd()

        print(
            Text.format(" Entrando no modo de filtragem ").center(
                RawTerminal.get_terminal_size(),
                Text.Token("═"),
            )
        )

        TkoFilterMode.deep_copy_and_change_dir()

        new_target_list: list[Path] = []

        for target in self.target_list:
            resolved: Path = target if target.is_absolute() else (old_dir / target)
            resolved = resolved.resolve()

            if resolved.exists():
                new_target_list.append(resolved)

        self.target_list = new_target_list
        
    def __print_diff(self):
        
        if self.wdir.get_solver().has_compile_error():
            executable, _ = self.wdir.get_solver().get_executable()
            print(executable.get_error_msg())
            return
        
        results = [unit.result for unit in self.wdir.get_unit_list()]
        if ExecutionResult.EXECUTION_ERROR not in results and ExecutionResult.WRONG_OUTPUT not in results:
            return
        
        if not self.param.compact:
            for elem in self.wdir.unit_list_resume():
                print(elem)

        if self.param.diff_count == DiffCount.NONE:
            return
        
        if self.param.diff_count == DiffCount.FIRST:
            # printing only the first wrong case
            wrong = [unit for unit in self.wdir.get_unit_list() if unit.result != ExecutionResult.SUCCESS][0]
            if self.param.diff_mode == DiffMode.DOWN:
                ud_diff_builder = DiffBuilderDown(RawTerminal.get_terminal_size(), wrong).to_insert_header()
                for line in ud_diff_builder.build_diff():
                    print(line)
            else:
                ss_diff_builder = DiffBuilderSide(RawTerminal.get_terminal_size(), wrong).to_insert_header(True)
                for line in ss_diff_builder.build_diff():
                    print(line)
            return

        if self.param.diff_count == DiffCount.ALL:
            for unit in self.wdir.get_unit_list():
                if unit.result != ExecutionResult.SUCCESS:
                    if self.param.diff_mode == DiffMode.DOWN:
                        ud_diff_builder = DiffBuilderDown(RawTerminal.get_terminal_size(), unit).to_insert_header()
                        for line in ud_diff_builder.build_diff():
                            print(line)
                    else:
                        ss_diff_builder = DiffBuilderSide(RawTerminal.get_terminal_size(), unit).to_insert_header(True)
                        for line in ss_diff_builder.build_diff():
                            print(line)

    def build_wdir(self):
        self.wdir_builded = True
        self.__remove_duplicates()
        self.__change_targets_to_filter_mode()
        try:
            self.wdir = Wdir(self.settings).set_curses(self.__curses_mode).set_lang(self.__lang).set_target_list(self.target_list).build().filter(self.param)
        except FileNotFoundError as e:
            if self.wdir.has_solver():
                executable, _ = self.wdir.get_solver().get_executable()
                executable.set_compile_error(str(e))
        return self.wdir

    def __missing_target(self) -> bool:
        if not self.wdir.has_solver() and not self.wdir.has_tests():
            if not self.__curses_mode:
                print(Text().addf("", "fail: ") + "Nenhum arquivo de código ou de teste encontrado.")
            return True
        return False
    
    def __list_mode(self):
        # list mode
        if not self.__eval_mode:
            print(Text.format("Nenhum arquivo de código encontrado. Listando casos de teste.").center(RawTerminal.get_terminal_size(), Text.Token("╌")), flush=True)
        print(self.wdir.resume_splitted())
        for line in self.wdir.unit_list_resume():
            print(line)

    def __free_run(self):        
        if self.__task is not None:
            if self.__rep:
                changes, size = self.store_exec_diff("---")
                self.__rep.logger.store(LogItemExec()
                    .set_mode(LogItemExec.Mode.FREE)
                    .set_key(self.__task.get_full_key() )
                    .set_size(changes, size))
        Free.free_run(self.wdir.get_solver(), show_compilation=False, to_clear=False, wait_input=False)

    def __create_opener_for_wdir(self) -> Opener:
        opener = Opener(self.settings)
        folders: list[Path] = []
        targets: list[Path] = [Path(".")]
        if self.target_list:
            targets = self.target_list

        for f in targets:
            if f.is_dir() and f not in folders:
                folders.append(f)
            else:
                folder = f.parent
                if folder not in folders:
                    folders.append(folder)
        opener.set_target(folders)
        if self.wdir.get_solver().args_list:
            solver_zero = self.wdir.get_solver().args_list[0]
            lang = solver_zero.name.split(".")[-1]
            opener.set_language(lang)
        return opener

    def __fill_task(self):
        task = Task()
        sources = self.wdir.get_source_list()
        solver = self.wdir.get_solver()

        
        if len(sources) > 0:
            target_path = os.path.abspath(sources[0])
        elif solver.args_list:
            target_path = os.path.abspath(self.wdir.get_solver().args_list[0])
        else:
            target_path = os.path.abspath(os.getcwd())

        if os.path.isfile(target_path):
            target_path = os.path.dirname(target_path)

        #break path in pieces rep_folder/database/key
        pieces = target_path.split(os.sep)
        if len(pieces) >= 3:
            task.set_key(pieces[-1])
            task.set_remote_name(pieces[-2])
            # task.set_rep_folder(os.sep.join(pieces[:-2]))
        elif len(pieces) == 2:
            task.set_key(pieces[-1])
            task.set_remote_name(pieces[-2])
            # task.set_rep_folder(os.sep)
        else:
            task.set_key(pieces[-1])
            task.set_remote_name("")
            # task.set_rep_folder(os.sep)

        self.__task = task
        self.__track_folder = None
        if self.__rep:
            self.__track_folder = self.__rep.paths.get_track_task_folder(task.get_full_key())

    def __run_test_on_curses(self):
        cdiff = Tester(self.settings, self.__rep, self.wdir, self.get_task())
        if self.__opener is not None:
            cdiff.set_opener(self.__opener)
        else:
            cdiff.set_opener(self.__create_opener_for_wdir())
        cdiff.set_autorun(self.__run_without_ask)
        cdiff.run()

    def get_rate(self) -> int:
        if self.__no_run:
            return 0
        correct = [unit for unit in self.wdir.get_unit_list() if unit.result == ExecutionResult.SUCCESS]
        percent = (len(correct) * 100) // len(self.wdir.get_unit_list())
        return percent

    def __run_tests_on_raw_term(self) -> int:
        if not self.__eval_mode:
            print(Text.format(" Testando o código com os casos de teste ").center(RawTerminal.get_terminal_size(), Text.Token("═")))
        percent = self.__run_all_tests_top_line()
        self.__print_diff()
        rate = self.get_rate()
        if self.__task is None or self.__track_folder is None:
            return rate

        exec_mode = LogItemExec.Mode.FULL if self.param.index is None else LogItemExec.Mode.LOCK
        exec_fail = LogItemExec.Fail.NONE
        if self.wdir.get_solver().has_compile_error():
            exec_fail = LogItemExec.Fail.COMP

        changes, size = self.store_exec_diff(str(rate))
        if self.__rep:
            self.__rep.logger.store(LogItemExec()
                .set_mode(exec_mode)
                .set_key(self.__task.get_full_key())
                .set_size(changes, size)
                .set_rate(rate)
                .set_fail(exec_fail))

        return percent

    def store_exec_diff(self, rate: str) -> tuple[bool, int]:
        tracker = Tracker()
        if self.__track_folder is not None:
            tracker.set_folder(self.__track_folder)
            tracker.set_files(self.wdir.get_solver().args_list)
            tracker.set_result(rate)
            has_changes, total_lines = tracker.store()
            return has_changes, total_lines
        return False, 0

    def __run_tests(self) -> int:
        if self.__task is None:
            self.__fill_task()
        if self.__curses_mode:
            self.__run_test_on_curses()
        else:
            return self.__run_tests_on_raw_term()

        return 0

    def __run_all_tests_top_line(self) -> int:
        if self.__eval_mode:
            key = self.get_task().get_full_key() if self.__task is not None else ""
            print(Text().addf("c", f"@{key:<{self.EVAL_MODE_PAD}} ").add(symbols.opening).add("[").add(self.wdir.resume_join()).add("]"), end="")
        else:
            print(Text().add(symbols.opening).add(self.wdir.resume_splitted()), end="")
        print(" [", end="")
        first = True
        execution_error = False
        for unit in self.wdir.get_unit_list():
            if first:
                first = False
            else:
                print(" ", end="")
            solver = self.wdir.get_solver()
            if self.__no_run or (execution_error and self.__abord_on_exec_error):
                unit.result = ExecutionResult.UNTESTED
            else:
                unit.result = UnitRunner.run_unit(solver, unit, timeout=self.__timeout)
                if unit.result == ExecutionResult.EXECUTION_ERROR:
                    execution_error = True
            print(Text() + ExecutionResult.get_symbol(unit.result), end="")
        print("] ", end="")
        if self.__show_track_info:
            if self.__rep is not None:
                logger = self.__rep.logger
                log_sort: LogSort | None = logger.tasks.task_dict.get(self.get_task().get_full_key(), None)
                if log_sort is not None:
                    log_resume = TaskResume(self.get_task().get_full_key()).from_log_sort(log_sort)
                    print(Text().addf("g", f"time:{log_resume.resume.minutes:.0f}, diff:{log_resume.resume.versions}, runs:{log_resume.resume.executions},").add(" "), end="", flush=True)

        if self.__no_run:
            percent: float = 0
        else:
            percent = self.get_rate()

        print(f"{percent:.0f}%")
        return round(percent)


import curses
import enum
import os
from typing import Callable, Optional

from tko.game.task import Task
from tko.play.floating_drop_down import FloatingDropDown
from tko.play.border import Border
from tko.play.floating import Floating
from tko.play.floating_drop_down import FloatingInputData
from tko.play.floating_manager import FloatingManager
from tko.play.fmt import Fmt
from tko.play.frame import Frame
from tko.play.images import compilling_image, executing_image, images, intro, random_get, success_image
from tko.play.keys import GuiActions, GuiKeys
from tko.play.opener import Opener
from tko.run.diff_builder import DiffBuilder
from tko.run.solver_builder import CompileError
from tko.run.unit import Unit
from tko.run.unit_runner import UnitRunner
from tko.run.wdir import Wdir
from tko.settings.settings import Settings
from tko.enums.execution_result import ExecutionResult
from tko.util.freerun import Free
from tko.util.text import  Text
from tko.util.symbols import symbols
from tko.enums.diff_mode import DiffMode
from tko.play.input_manager import InputManager
from tko.run.diff_builder_down import DiffBuilderDown
from tko.run.diff_builder_side import DiffBuilderSide
from tko.play.tracker import Tracker
from tko.util.raw_terminal import RawTerminal
from tko.settings.repository import Repository
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_move import LogItemMove
from tko.util.symbols import symbols
from tko.logger.log_item_self import LogItemSelf
from tko.play.floating_grade import FloatingGrade

class SeqMode(enum.Enum):
    intro = 0
    select = 1
    running = 2
    finished = 3

class Tester:
    def __init__(self, settings: Settings, rep: Repository | None, wdir: Wdir, task: Task):
        self.results: list[tuple[ExecutionResult, int]] = []
        self.wdir = wdir
        self.rep: Repository | None = rep
        self.unit_list = [unit for unit in wdir.get_unit_list()] # unit list to be consumed
        self.exit = False
        self.task = task
        self.diff_first_line = 1000   # index of first line to show
        self.length = 1  # length of diff
        self.space = 0  # dy space for draw
        self.mode: SeqMode = SeqMode.intro

        self.settings = settings
        self.app = settings.app
        self.borders = Border(settings.app)

        self.locked_index: bool = False
        self.focused_index = 0
        self.resumes: list[str] = []

        self.fman = FloatingManager()
        self.opener: Opener | None = None
        self.dummy_unit = Unit()
        if self.rep:
            self.rep.logger.store(
                LogItemMove().set_mode(LogItemMove.Mode.PICK).set_key(self.task.get_full_key())
            )


    def set_opener(self, opener: Opener):
        self.opener = opener
        self.opener.set_fman(self.fman)
        return self

    def set_autorun(self, value:bool):
        if value:
            self.mode = SeqMode.running
        return self

    def set_exit(self):
        self.exit = True
        return self

    @staticmethod
    def print_centered_image(image: str, color: str, clear: bool = False, align: str = "."):
        dy, dx = Fmt.get_size()
        lines = image.splitlines()[1:]
        init_y = 4
        if align == "v":
            init_y = dy - len(lines) - 1
        for i, line in enumerate(lines):
            info = Text().addf(color, line).center(dx - 2, Text.Token(" ", " "))
            if clear:
                Fmt.write(i + init_y, 1, Text().add(" " * info.len()))
            else:
                Fmt.write(i + init_y, 1, Text().addf(color, line).center(dx - 2, Text.Token(" ", " ")))

    def show_success(self):
        if self.settings.app.get_use_images():
            out = random_get(images, self.get_folder(), "static")
        else:
            out = random_get(success_image, self.get_folder(), "static")
        Tester.print_centered_image(out, "g")
        
    def show_compilling(self, clear: bool = False):
        out = random_get(compilling_image, self.get_folder(), "random")
        Tester.print_centered_image(out, "y", clear)

    @staticmethod
    def show_executing(clear: bool = False):
        out = executing_image
        Tester.print_centered_image(out, "y", clear, "v")

    def get_folder(self) -> str:
        folder = self.task.get_workspace_folder()
        if folder is None:
            raise Warning("Warning: Task folder não encontrado")
        return os.path.basename(folder)

    def get_focused_unit(self) -> Unit:
        if not self.wdir.has_tests():
            return self.dummy_unit
        if len(self.results) != 0:
            _, index = self.results[self.focused_index]
            unit = self.wdir.get_unit(index)
            return unit
        return self.wdir.get_unit(self.focused_index)

    @staticmethod
    def get_token(result: ExecutionResult) -> Text.Token:
        if result == ExecutionResult.SUCCESS:
            return Text.Token(ExecutionResult.get_symbol(ExecutionResult.SUCCESS).text, "G")
        elif result == ExecutionResult.WRONG_OUTPUT:
            return Text.Token(ExecutionResult.get_symbol(ExecutionResult.WRONG_OUTPUT).text, "R")
        elif result == ExecutionResult.COMPILATION_ERROR:
            return Text.Token(ExecutionResult.get_symbol(ExecutionResult.UNTESTED).text, "W")
        elif result == ExecutionResult.EXECUTION_ERROR:
            return Text.Token(ExecutionResult.get_symbol(ExecutionResult.EXECUTION_ERROR).text, "Y")
        else:
            return Text.Token(ExecutionResult.get_symbol(ExecutionResult.UNTESTED).text, "W")

    def store_version(self, result: str) -> tuple[bool, int]:
        if self.rep is None:
            return False, 0
        track_folder = self.rep.paths.get_track_task_folder(self.task.get_full_key())
        tracker = Tracker()
        tracker.set_folder(track_folder)
        tracker.set_files(self.wdir.get_solver().args_list)
        tracker.set_result(result)
        has_changes, total_lines = tracker.store()
        return has_changes, total_lines


    def process_one(self):

        if self.mode != SeqMode.running:
            return

        solver = self.wdir.get_solver()

        if solver.has_compile_error():
            mode = LogItemExec.Mode.LOCK if self.locked_index else LogItemExec.Mode.FULL
            changes, total_lines = self.store_version("0")
            if self.rep:
                self.rep.logger.store(
                    LogItemExec().set_key(self.task.get_full_key()).set_mode(mode).set_fail(LogItemExec.Fail.COMP).set_size(changes, total_lines)
                )

            self.mode = SeqMode.finished
            while len(self.unit_list) > 0:
                index = len(self.results)
                self.unit_list = self.unit_list[1:]
                self.results.append((ExecutionResult.COMPILATION_ERROR, index))
            return
        
        if self.locked_index or not self.wdir.has_tests():
            self.mode = SeqMode.finished
            unit = self.get_focused_unit()
            unit.result = UnitRunner.run_unit(solver, unit, self.settings.app.timeout)

            rate = 100 if unit.result == ExecutionResult.SUCCESS else 0
            changes, total_lines = self.store_version(str(rate))
            if self.rep:
                self.rep.logger.store(
                    LogItemExec().set_key(self.task.get_full_key()).set_mode(LogItemExec.Mode.LOCK).set_rate(rate).set_size(changes, total_lines)
                )
            return

        if self.wdir.has_tests():
            index = len(self.results)
            unit = self.unit_list[0]
            self.unit_list = self.unit_list[1:]
            unit.result = UnitRunner.run_unit(solver, unit, self.settings.app.get_timeout())
            self.results.append((unit.result, index))
            self.focused_index = index
            if unit.result == ExecutionResult.EXECUTION_ERROR:
                self.mode = SeqMode.finished
                while len(self.unit_list) > 0:
                    index = len(self.results)
                    self.unit_list = self.unit_list[1:]
                    self.results.append((ExecutionResult.EXECUTION_ERROR, index))

        if len(self.unit_list) == 0:
            self.mode = SeqMode.finished
            self.focused_index = 0

            done_list: list[tuple[ExecutionResult, int]] = []
            fail_list: list[tuple[ExecutionResult, int]] = []
            for data in self.results:
                unit_result, _ = data
                if unit_result != ExecutionResult.SUCCESS:
                    fail_list.append(data)
                else:
                    done_list.append(data)
            self.results = fail_list + done_list
            percent: int = (100 * len(done_list)) // len(self.results)
            self.task.info.rate = percent
            mode = LogItemExec.Mode.LOCK if self.locked_index else LogItemExec.Mode.FULL
            changes, total_lines = self.store_version(str(percent))
            if self.rep:
                self.rep.save_config()
                self.rep.logger.store(
                    LogItemExec().set_key(self.task.get_full_key()).set_mode(mode).set_rate(percent).set_size(changes, total_lines)
                )


    def build_top_line_header(self, frame_dx : int) -> Text:
        activity_color = "C"
        solver_color = "W"
        sources_color = "Y"
        running_color = "R"

        # building activity
        activity = Text().add(self.borders.border(activity_color, self.get_folder()))

        # building solvers
        solvers = Text()
        if len(self.get_solver_names()) > 1:
            solvers.add(Text().add(self.borders.round_l("R")).addf("R", f"{GuiActions.tab}").add(self.borders.sharp_r("R")))
        for i, solver in enumerate(self.get_solver_names()):
            if len(self.get_solver_names()) > 1:
                solvers.add(" ")
            color = solver_color
            if i == self.task.main_idx:
                color = "G"
            solvers.add(self.borders.border(color, solver))
        
        # replacing with count if running
        done = len(self.results)
        full = len(self.wdir.get_unit_list())
        count_missing = Text().add(self.borders.border(running_color, f"({done}/{full})"))
        if self.mode == SeqMode.running:
            if  self.locked_index:
                solvers = Text().add(self.borders.border("R", "Executando atividade travada"))
            else:
                solvers = count_missing

        # building sources
        source_names = Text().add(", ").join([Text().addf(sources_color, f"{name[0]}({name[1]})") for name in self.wdir.sources_names()])
        if self.wdir.has_tests():
            sources = Text().add(self.borders.round_l(sources_color)).add(source_names).add(self.borders.round_r(sources_color))
        else:
            sources = Text().add(self.borders.border("R", "Nenhum teste cadastrado"))

        # merging activity, solvers and sources in header
        delta = frame_dx - solvers.len()
        left = 1
        right = 1
        if delta > 0:
            delta_left = delta // 2
            left = max(1, delta_left - activity.len())
            delta_right = delta - delta_left
            right = max(1, delta_right - sources.len())
        filler = "─" if self.wdir.has_tests() else " "
        return Text().add(activity).add(filler * left).add(solvers).add(filler * right).add(sources)

    def build_unit_list(self, frame: Frame) -> Text:
        done_list = self.results
        if len(done_list) > 0 and self.locked_index:
            _, index = done_list[self.focused_index]
            done_list[self.focused_index] = (self.get_focused_unit().result, index)
        todo_list: list[tuple[ExecutionResult, int]] = []
        i = len(done_list)
        for _ in self.unit_list:
            if self.locked_index and i == self.focused_index:
                todo_list.append((self.wdir.get_unit(self.focused_index).result, i))
            else:
                todo_list.append((ExecutionResult.UNTESTED, i))
            i += 1

        i = 0
        show_focused_index = not self.wdir.get_solver().has_compile_error() and not self.mode == SeqMode.intro and not self.is_all_right()
        
        output = Text()
        output.add(self.get_fixed_arrow())


        for unit_result, index in done_list + todo_list:
            foco = i == self.focused_index
            token = Tester.get_token(unit_result)
            extrap = self.borders.round_l(token.fmt)
            extras = self.borders.round_r(token.fmt)
            if foco and show_focused_index:
                # token.fmt = token.fmt.lower() + ""
                extrap = Text.Token("(")
                extras = Text.Token(")")
                # extrap = Text.Token(" ") #self.borders.roundL("")
                # extras = Text.Token(" ") #self.borders.roundR("")
            if self.locked_index and not foco:
                output.add("  ").addf(token.fmt.lower(), str(index).zfill(2)).addf(token.fmt.lower(), token.text).add(" ")
            else:
                output.add(" ").add(extrap).addf(token.fmt, str(index).zfill(2)).add(token).add(extras)
            i += 1

        size = 6
        to_remove = 0
        index = self.focused_index
        dx = frame.get_dx()
        while (index + 4) * size - to_remove >= dx:
            to_remove += size
        output.data = output.data[:3 * 6] + output.data[3 * 6 + to_remove:]
        return output

    def get_fixed_arrow(self) -> Text:
        
        output = Text()
        # diff
        diff_text = self.get_diff_symbol()
        output.addf("B", f" {GuiKeys.diff} {diff_text} ")
        # if self.settings.app.has_borders():
        #     output.addf("Mb", symbols.sharp_r.text)
        # else:
        #     output.addf("B", " ")
        # timer
        # timer = symbols.timer.text
        # value = str(self.settings.app.get_timeout())
        # if value == "0":
        #     value = symbols.infinity.text
        # output.addf("M", f"{timer}l {value}  ")
        color = "R" if self.locked_index else "G"
        if self.settings.app.get_use_borders():
            output.addf(color + "b", symbols.sharp_r.text)
        else:
            output.addf("B", " ")

        # travar
        free = symbols.locked_free.text
        locked = symbols.locked_locked.text
        symbol = locked if self.locked_index else free
        output.addf(color, f" {GuiKeys.lock} {symbol} ").add(self.borders.sharp_r(color))

        return output

    def draw_top_bar_content(self, frame : Frame):
        if not self.wdir.has_tests():
            return
        value = self.get_focused_unit()
        info = Text()
        if self.wdir.get_solver().has_compile_error():
            info = self.borders.border("R", "Erro de compilação")
        elif not self.is_all_right() and not self.mode == SeqMode.intro:
            info = value.str(pad = False)
            # if self.locked_index:
            #     info = self.borders.border(focused_unit_color, info.get_text())
        frame.write(0, 0, Text().add(info).center(frame.get_dx()))

    def draw_top_bar(self):
        # construir mais uma solução
        _, cols = Fmt.get_size()
        size = 1
        if self.wdir.has_tests():
            size = 3
        frame = Frame(0, 0).set_size(size, cols)
        if not self.wdir.has_tests():        
            frame.set_border_none()
        frame.set_header(self.build_top_line_header(frame.get_dx()))
        self.draw_top_bar_content(frame)
        frame.set_footer(self.build_unit_list(frame), "")
        frame.draw()

    def get_diff_symbol(self) -> str:
        if self.settings.app.get_diff_mode() == DiffMode.DOWN:
            return symbols.diff_down.text
        return symbols.diff_left.text

    def make_bottom_line(self) -> list[Text]:
        cmds: list[Text] = []

        text = f"{GuiActions.pallete} [{GuiKeys.palette}]"
        cmds.append(self.borders.border("C", f"Tela Anterior[Esc]"))
        cmds.append(self.borders.border("C", text))
        if self.opener is not None:
            cmds.append(self.borders.border("C", f"{GuiActions.edit} [{GuiKeys.edit}]"))
        cmds.append(self.borders.border("G", f"{GuiActions.evaluate_tester} [{GuiKeys.evaluate} | {symbols.newline.text}]"))
        cmds.append(self.borders.border("G", f"{GuiActions.execute_tester} [{GuiKeys.execute} | ←]"))
        limite = f"{GuiActions.time_limit} {self.get_time_limit_symbol()} [{GuiKeys.limite}]"
        cmds.append(self.borders.border("Y", limite))

        return cmds

    def show_bottom_line(self):
        lines, cols = Fmt.get_size()
        out = Text().add(" ").join(self.make_bottom_line())
        Fmt.write(lines - 1, 0, out.center(cols, Text.Token(" ")))

    def is_all_right(self):
        if self.locked_index or len(self.results) == 0:
            return False
        if not self.mode == SeqMode.finished:
            return False
        for result, _ in self.results:
            if result != ExecutionResult.SUCCESS:
                return False
        return True

    def draw_main(self):
        unit = self.get_focused_unit()
        lines, cols = Fmt.get_size()
        if self.wdir.has_tests():
            y_out = 2
            self.space = lines - 4
        else:
            y_out = 0
            self.space = lines - 2
        frame = Frame(y_out, -1).set_inner(self.space, cols)

        if self.is_all_right():
            self.show_success()
            return
        
        diff_builder = DiffBuilder(cols)
        diff_builder.set_curses()

        if self.wdir.get_solver().has_compile_error():
            executable, _ = self.wdir.get_solver().get_executable()
            received = executable.get_error_msg().get_str()
            line_list = [Text().add(line) for line in received.splitlines()]
        elif self.settings.app.get_diff_mode() == DiffMode.DOWN or not self.wdir.has_tests():
            ud_diff = DiffBuilderDown(cols, unit)
            line_list = ud_diff.build_diff()
        else:
            ss_diff = DiffBuilderSide(cols, unit)
            line_list = ss_diff.build_diff()

        self.length = max(1, len(line_list))

        if self.length - self.diff_first_line < self.space:
            self.diff_first_line = max(0, self.length - self.space)

        if self.diff_first_line >= self.length:
            self.diff_first_line = self.length - 1

        if self.diff_first_line < self.length:
            line_list = line_list[self.diff_first_line:]
        for i, line in enumerate(line_list):
            frame.write(i, 0, Text().add(line))
        return

    def get_solver_names(self):
        return sorted(self.wdir.solvers_names())
    
    def main(self, scr: curses.window) -> Optional[Callable[[], bool]]:
        InputManager.fix_esc_delay()
        curses.curs_set(0)  # Esconde o cursor
        Fmt.init_colors()  # Inicializa as cores
        Fmt.set_scr(scr)  # Define o scr como global
        while not self.exit:
            Fmt.clear()
            if self.mode == SeqMode.running:
                if self.wdir.get_solver().not_compiled():
                    self.draw_top_bar()
                    self.show_bottom_line()
                    self.show_compilling()
                    Fmt.refresh()
                    try:
                        self.wdir.get_solver().prepare_exec()
                    except CompileError as e:
                        self.fman.add_input(Floating().bottom().right().set_error().put_text(e.message))
                        self.mode = SeqMode.finished
                    Fmt.clear()
                    self.draw_top_bar()
                    self.show_bottom_line()
                    Fmt.refresh()
                self.process_one()
            self.draw_top_bar()

            if self.mode == SeqMode.intro:
                Tester.print_centered_image(random_get(intro, self.get_folder()), "y")
            else:
                self.draw_main()
            
            self.show_bottom_line()

            if self.fman.has_floating():
                self.fman.draw_warnings()

            if self.mode == SeqMode.running:
                Fmt.refresh()
                continue

            if self.fman.has_floating():
                self.fman.draw()

            # o input tem que ser depois do draw para mostrar o floating
            value = InputManager.get_and_remap_keys(scr, self.app)
            
            if self.fman.has_floating():
                # if consumed, value becomes -1
                value = self.fman.process_input(value)

            fn_exec = self.process_key(value)
            if fn_exec is not None:
                return fn_exec
        return None

    def run_exec_mode(self) -> Callable[[], bool]:
        self.mode = SeqMode.running
        if self.wdir.is_autoload():
            self.wdir.autoload()
            self.wdir.get_solver().set_main(self.get_solver_names()[self.task.main_idx])
        self.mode = SeqMode.finished
        changes, total_lines = self.store_version("---")
        if self.rep:
            self.rep.logger.store(
                LogItemExec().set_key(self.task.get_full_key()).set_mode(LogItemExec.Mode.FREE).set_size(changes, total_lines)
            )
        header = self.build_top_line_header(RawTerminal.get_terminal_size())
        return lambda: Free.free_run(self.wdir.get_solver(), show_compilation=True, to_clear=True, wait_input=True, header=header)

    def run_test_mode(self):
        self.mode = SeqMode.running
        if self.wdir.is_autoload():
            self.wdir.autoload() # reload sources and solvers
        self.wdir.build() # reload tests

        Fmt.clear()

        solver_names = self.get_solver_names()
        solver_size = len(solver_names)
        index = self.task.main_idx
        solver_selected = solver_names[index % solver_size]
        self.wdir.get_solver().set_main(solver_selected).reset() # clear old compilation
        
        if self.locked_index:
            for i in range(len(self.results)):
                _, index = self.results[i]
                self.results[i] = (ExecutionResult.UNTESTED, index)
        else:
            self.focused_index = 0
            self.results = []
            self.unit_list = [unit for unit in self.wdir.get_unit_list()]


    def go_left(self):
        if self.mode == SeqMode.intro:
            self.mode = SeqMode.select
        if self.mode == SeqMode.finished:
            self.mode = SeqMode.select
        if self.locked_index:
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("←\nAtividade travada\nAperte {} para destravar".format(GuiKeys.lock)))
            return
        if not self.wdir.get_solver().has_compile_error():
            self.focused_index = max(0, self.focused_index - 1)
            self.diff_first_line = 1000

    def go_right(self):
        if self.mode == SeqMode.intro:
            self.mode = SeqMode.select
            self.focused_index = 0
            return
        if self.mode == SeqMode.finished:
            self.mode = SeqMode.select
        if self.locked_index:
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("→\nAtividade travada\nAperte {} para destravar".format(GuiKeys.lock)))
            return
        if not self.wdir.get_solver().has_compile_error():
            self.focused_index = min(len(self.wdir.get_unit_list()) - 1, self.focused_index + 1)
            self.diff_first_line = 1000

    def go_down(self):
        if self.mode == SeqMode.intro:
            self.mode = SeqMode.select
        self.diff_first_line += 1

    def go_up(self):
        if self.mode == SeqMode.intro:
            self.mode = SeqMode.select
        self.diff_first_line = max(0, self.diff_first_line - 1)

    def change_main(self):
        if len(self.get_solver_names()) == 1:
            self.fman.add_input(
                Floating().bottom().right().set_warning()
                .put_text("Seu projeto só tem um arquivo de solução.")
                .put_text("Essa funcionalidade troca qual dos arquivos")
                .put_text("de solução será o principal.")
            )
            return
        self.task.main_idx = (self.task.main_idx + 1) % len(self.get_solver_names())

    def lock_unit(self):
        self.locked_index = not self.locked_index
        if self.mode == SeqMode.intro:
            self.mode = SeqMode.select
        if self.locked_index:
            for i in range(len(self.results)):
                _, index = self.results[i]
                self.results[i] = (ExecutionResult.UNTESTED, index)

    def get_time_limit_symbol(self):
        if self.settings.app.get_timeout() == 0:
            return symbols.infinity.text
        return str(self.settings.app.get_timeout())

    def change_limit(self):
            valor = self.settings.app.get_timeout()
            if valor == 0:
                valor = 1
            else:
                valor *= 2
            if valor >= 5:
                valor = 0
            self.settings.app.set_timeout(valor)
            self.settings.save_settings()

    def process_key(self, key: int) -> Optional[Callable[[], bool]]:
        if key == ord('q'):
            self.set_exit()
        elif key == curses.KEY_EXIT:
            if self.locked_index:
                self.locked_index = False
            else:
                self.set_exit()
        elif key == curses.KEY_LEFT:
            self.go_left()
        elif key == curses.KEY_RIGHT:
            self.go_right()
        elif key == curses.KEY_DOWN:
            self.go_down()
        elif key == curses.KEY_UP:
            self.go_up()
        elif key == ord(GuiKeys.toggle_main):
            self.change_main()
        elif key == ord(GuiKeys.execute) or key == curses.KEY_BACKSPACE:
            return self.run_exec_mode()
        elif key == ord(GuiKeys.evaluate) or key == ord('\n'):
            self.run_test_mode()
        elif key == ord(GuiKeys.lock):
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Função de travamento {}".format("ligada" if not self.locked_index else "desligada")))
            self.lock_unit()
        elif key == ord(GuiKeys.edit):
            if self.opener is not None:
                self.opener.load_folders_and_open()
        elif key == ord(GuiKeys.self_evaluate):
            self.self_evaluate()
        elif key == ord(GuiKeys.limite):
            self.change_limit()
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Limite de execução alterado para {}".format(self.get_time_limit_symbol())))
            self.settings.save_settings()
        elif key == ord(GuiKeys.diff):
            self.settings.app.toggle_diff()
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Modo de Diff alterado para {}".format(self.settings.app.get_diff_mode().value)))
            self.settings.save_settings()
        elif key == ord(GuiKeys.borders):
            self.settings.app.toggle_borders()
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Modo de Bordas alterado para {}".format("ligado" if self.settings.app.get_use_borders() else "desligado")))
            self.settings.save_settings()
        elif key == ord(GuiKeys.images):
            self.settings.app.toggle_images()
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Modo de Imagens alterado para {}".format("ligado" if self.settings.app.get_use_images() else "desligado")))
            self.settings.save_settings()
        elif key == ord(GuiKeys.palette):
            self.command_pallete()
        elif key != -1 and key != curses.KEY_RESIZE:
            self.fman.add_input( Floating().bottom().right().set_error().put_text(f"Tecla char:{chr(key)}, code:{key}, não reconhecida") )
        return None

    def self_evaluate(self):
        if self.rep is None:
            self.fman.add_input(Floating().bottom().right().set_warning().put_text("Nenhum repositório de logs encontrado."))
            return
        logger = self.rep.logger
        self.fman.add_input(
            FloatingGrade(self.task, lambda task: logger.store(LogItemSelf().set_task(task)))
        )

    def command_pallete(self):
        options: list[FloatingInputData] = []

        def icon(value: bool):
            return "✓" if value else "✗"
        
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mudar arquivo {y} de execução", symbols.action, "principal"),
                self.change_main,
                "TAB"
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mudar modo {y}}", self.get_diff_symbol(), "Diff"),
                self.app.toggle_diff,
                GuiKeys.diff
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mudar {y} de tempo de execução: {r}", symbols.action, "Limite", self.get_time_limit_symbol()),
                self.change_limit,
                GuiKeys.limite
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Testar {y} os casos ou apenas o selecionado", icon(not self.locked_index), "Todos"),
                self.lock_unit,
                GuiKeys.lock
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_borders()), "Bordas"),
                self.app.toggle_borders,
                GuiKeys.borders
            )
        )
        
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_images()), "Imagens"),
                self.app.toggle_images, 
                GuiKeys.images
            )
        )

                # self evaluate
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: Auto {y} método de estudo", symbols.action, "Avaliar"),
                self.self_evaluate,
                GuiKeys.self_evaluate
            ).set_exit_on_action(True)
        )

        self.fman.add_input(
            FloatingDropDown().set_floating(
                        Floating().set_text_ljust()
                        .set_header(" Selecione uma ação da lista ")
                        .set_footer(" Use Enter para aplicar e Esc para Sair ")
                      )
                      .set_options(options)
                      .set_exit_on_enter(False)
        )



    def run(self):
        while True:
            free_run_fn = curses.wrapper(self.main) # type: ignore
            if free_run_fn is None:
                if self.rep:
                    self.rep.logger.store(
                            LogItemMove().set_mode(LogItemMove.Mode.BACK).set_key(self.task.get_full_key())
                        )
                break
            else:
                while True:
                    try:
                        repeat = free_run_fn()
                        if not repeat:
                            break
                    except CompileError as e:
                        self.mode = SeqMode.finished
                        if self.rep:
                            self.rep.logger.store(
                                            LogItemExec().set_key(self.task.get_full_key()).set_mode(LogItemExec.Mode.FREE).set_fail(LogItemExec.Fail.COMP)
                                        )
                        print(e)
                        input("Pressione enter para continuar")
                        break



from __future__ import annotations
import datetime as dt
import enum


class Delta:
    format = '%Y-%m-%d %H:%M:%S'
    def __init__(self):
        self.mode: Delta.Mode | None = None
        self.datetime: dt.datetime = dt.datetime.fromordinal(1)
        self.elapsed: dt.timedelta = dt.timedelta()
        self.accumulated: dt.timedelta = dt.timedelta()

    def __str__(self) -> str:
        return f"datetime:{Delta.encode_format(self.datetime)}, elapsed:{Delta.encode_timedelta(self.elapsed)}, acc:{Delta.encode_timedelta(self.accumulated)}"


    @staticmethod
    def encode_timedelta(elapsed: dt.timedelta) -> str:
        total_seconds = elapsed.total_seconds()
        minutes = round(total_seconds // 60)
        seconds = round(total_seconds) % 60
        return f'{minutes:03d}:{seconds:02d}'
    
    def create(self, mode: Mode, last_item: Delta | None, datetime: dt.datetime) -> Delta:
        self.mode = mode
        if mode.action == Delta.Mode.Action.without_inc_time:
            return self.__create_without_inc_time(last_item, datetime)
        elif mode.action == Delta.Mode.Action.incrementing_time:
            return self.__create_incrementing_time(last_item, datetime)
        elif mode.action == Delta.Mode.Action.with_time_threshold:
            return self.__create_with_time_threshold(last_item, datetime, mode.minutes_limit)
        else:
            raise ValueError(f'Unknown action {mode.action}')

    def __create_without_inc_time(self, last_item: Delta | None, datetime: dt.datetime) -> Delta:
        if last_item is None:
            self.datetime = datetime
            self.elapsed = dt.timedelta(seconds=0)
            return self

        self.datetime = datetime
        if last_item.datetime < datetime:
            self.elapsed = datetime - last_item.datetime
        self.accumulated = last_item.accumulated
        return self

    def __create_incrementing_time(self, last_item: Delta | None, datetime: dt.datetime) -> Delta:
        self.__create_without_inc_time(last_item, datetime)
        if last_item is not None and self.elapsed > dt.timedelta(seconds=0):
            self.accumulated += self.elapsed
        return self

    def __create_with_time_threshold(self, last_item: Delta | None, datetime: dt.datetime, minutes_limit: int) -> Delta:
        self.__create_without_inc_time(last_item, datetime)
        if last_item is not None and self.elapsed > dt.timedelta(seconds=0) and self.elapsed < dt.timedelta(minutes=minutes_limit):
            self.accumulated += self.elapsed
        return self

    @staticmethod
    def now() -> tuple[str, dt.datetime]:
        now = dt.datetime.now()
        return now.strftime(Delta.format), now

    @staticmethod
    def decode_format(value: str) -> dt.datetime:
        return dt.datetime.strptime(value, Delta.format)
    
    @staticmethod
    def encode_format(value: dt.datetime) -> str:
        return dt.datetime.strftime(value, Delta.format)
    
    @staticmethod
    def format_h_min(hours: float) -> str:
        if hours < 0:
            return "00:00"
        h = int(hours)
        m = int((hours - h) * 60)
        return f"{h:02d}h {m:02d}m"

    @staticmethod
    def week_day(day: str) -> str:
        date = dt.datetime.strptime(day, '%Y-%m-%d')
        return date.strftime('%A')

    @staticmethod
    def next_day(day: str) -> str:
        date = dt.datetime.strptime(day, '%Y-%m-%d')
        return (date + dt.timedelta(days=1)).strftime('%Y-%m-%d')

    class Mode:        
        def __init__(self, action: Action, minutes_limit: int = 60):
            self.action = action
            self.minutes_limit = minutes_limit

        def __str__(self) -> str:
            return f'{self.action.name} {self.minutes_limit}'

        class Action(enum.Enum):
            without_inc_time = 1
            incrementing_time = 2
            with_time_threshold = 3


class GuiActions:
    tab = "↔TAB"
    leave   = "Sair"
    edit = "Ver tarefa"
    show_time = "Tempo"
    filter = "Filtrar"
    principal = "Principal"
    github = "Web   "
    help = "Ajuda"
    download = "Baixar"
    activate = "Ativar"
    navigate = "←↓→"
    uncheck = "Desmarcar"
    search = "Busca"
    pallete = "Paleta"
    grade = "Auto Avaliação"
    # tester frame
    time_limit = "Limite de tempo"
    evaluate_tester = "Testar"
    execute_tester  = "Executar"
    calibrate = "Calibrar teclas"
    draft = "Rascunho"
    all = "Todos"

class GuiKeys:
    calibrate = "c"

    inbox = "1"
    all_tasks = "2"
    
    panel_graph = "3"
    panel_logs = "4"
    panel_skills = "5"
    panel_help = "6"

    show_duration = "T"

    palette = "p"
    images = "I"
    down_task = "b"
    activate = "\n"
    reload_game = "R"
    
    create_draft = "r"
    set_lang_drafts = "L"
    open_url = "u"
    key_quit = "q"
    edit= "v"
    colors = "C"
    borders = "B"
    search = "/"
    self_evaluate = "a"
    delete_folder = "D"
    diff = "d"
    unfold_patch = "V"
    execute = "e"
    evaluate = "t"
    toggle_main = "\t"
    quit   = "q"
    lock = "u"
    limite  = "l"
    expand = ">"
    collapse = "<"


from tko.play.flags import Flags
from tko.play.floating_drop_down import FloatingDropDown
from tko.play.floating_drop_down import FloatingInputData
from tko.util.text import Text
from tko.util.symbols import symbols
from tko.play.keys import GuiKeys
from tko.play.play_actions import PlayActions
from tko.play.floating_calibrate import FloatingCalibrate
from tko.play.floating import Floating

class PlayPalette:
    def __init__(self, actions: PlayActions):
        self.actions = actions
        self.app = self.actions.settings.app
        self.fman = self.actions.fman
        self.gui = self.actions.gui

    def command_pallete(self):
        options: list[FloatingInputData] = []

        def icon(value: bool):
            return Text.Token("✓", "g") if value else Text.Token("✗", "r")
        
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: {y} para o repositório", symbols.action, "Baixar"),
                self.actions.down_remote_task,
                GuiKeys.down_task
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: Abrir {y} com a descrição", symbols.action, "URL"),
                self.actions.open_link,
                GuiKeys.open_url
            ).set_exit_on_action(True)
        )

        # self evaluate
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: Auto {y} método de estudo", symbols.action, "Avaliar"),
                self.actions.self_evaluate,
                GuiKeys.self_evaluate
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: {y} pasta", symbols.action, "Apagar"),
                self.actions.delete_folder_ask,
                GuiKeys.delete_folder
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", symbols.action, "Ajuda"),
                lambda: Flags.panel.set_value(Flags.panel_help),
                GuiKeys.panel_help
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_borders()), "Bordas"),
                self.app.toggle_borders,
                GuiKeys.borders
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_images()), "Imagens"),
                self.app.toggle_images, 
                GuiKeys.images
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y} decorrido", icon(Flags.show_time.is_true()), "Tempo"),
                lambda: Flags.show_time.toggle(),
                GuiKeys.show_duration
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mudar {y} de programação", symbols.action, "Linguagem"),
                self.gui.language.set_language,
                GuiKeys.set_lang_drafts
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} {y} as teclas direcionais", symbols.action, "Calibrar"),
                lambda: self.fman.add_input(FloatingCalibrate(self.actions.settings)),
                GuiKeys.calibrate
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Criar {y} na pasta local", symbols.action, "Rascunho"),
                self.actions.create_draft,
                GuiKeys.create_draft
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} {y} pasta de rascunhos", symbols.action, "Reload"),
                self.actions.reload_game,
                GuiKeys.reload_game
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Ver {y} da tarefa ", symbols.action, "versões"),
                self.actions.open_versions,
                GuiKeys.unfold_patch
            ).set_exit_on_action(True)
        )

        self.fman.add_input(
            FloatingDropDown().set_floating(
                    Floating().top()
                    .set_text_ljust()
                    .set_footer(" Use Enter para aplicar e Esc para Sair ")
                    .set_header(" Selecione uma ação da lista ")
            )
            .set_options(options)
            .set_exit_on_enter(False)
        )


from tko.play.floating import Floating, FloatingABC
from tko.util.symbols import symbols
from tko.util.text import Text
from typing import Callable

import curses

class FloatingInputData:
    def __init__(self, label: Callable[[], Text], action: Callable[[], None], shortcut: str = ""):
        self.label = label
        self.action = action
        self.shortcut = shortcut
        self.exit_on_action = False

    def set_exit_on_action(self, value: bool):
        self.exit_on_action = value
        return self


class FloatingDropDown(FloatingABC):
    def __init__(self):
        self.floating = Floating()
        self.index = 0
        self.options: list[FloatingInputData] = []
        self.floating.frame.set_border_color("m")
        self.exit_on_action = True

        self.search_text: list[str] = []

    def set_floating(self, floating: Floating):
        self.floating = floating
        return self

    # @override
    def is_enable(self) -> bool:
        return self.floating.is_enable()
    
    # @override
    def draw(self):
        self.update_content()
        self.floating.draw()

    def calc_dy_dx(self):
        dy, dx = self.floating.calc_dy_dx()
        dy += len(self.options) + 2
        for option in self.options:
            dx = max(dx, len(option.label()))
        return dy, dx

    def set_exit_on_enter(self, value: bool):
        self.exit_on_action = value
        return self

    def match_search(self, index: int):
        return "".join(self.search_text) in self.options[index].label().get_str().lower()

    def next_option(self):
        if not self.match_search(self.index):
            self.index = 0
        steps = len(self.options)
        index = self.index
        while steps > 0:
            index = (index + 1) % len(self.options)
            if self.match_search(index):
                self.index = index
                return
            steps -= 1

    def prev_option(self):
        if not self.match_search(self.index):
            self.index = 0
        steps = len(self.options)
        index = self.index
        while steps > 0:
            index = (index - 1) % len(self.options)
            if self.match_search(index):
                self.index = index
                return
            steps -= 1

    def update_content(self):
        content = self.floating.content
        content.clear()
        content.append(Text.format("Busca: ") + "".join(self.search_text) + symbols.cursor)

        # options: list[Text] = []
        _, dx = self.calc_dy_dx()
        for i, option in enumerate(self.options):
            if not self.match_search(i):
                continue
            text = Text().add(option.label()).ljust(dx)
            if option.shortcut != "":
                if len(option.shortcut) > 1:
                    text.add(" " + option.shortcut)
                else:
                    text.add(f" [{option.shortcut}]")
            else:
                text.add("    ")
            fmt = "M" if i == self.index else ""
            text.set_background(fmt)
            content.append(text)
        

        # y = 1
        # for line in self.floating.content + options:
        #     x = 0
        #     if self.floating.centralize_text:
        #         x = (self.floating.frame.get_dx() - line.len()) // 2
        #     self.floating.frame.write(y, x, line)
        #     y += 1

        return self

    def set_options(self, options: list[FloatingInputData]):
        self.options = options
        return self

    def set_default_index(self, index: int):
        self.index = index
        return self

    def update_index(self):
        for i, _ in enumerate(self.options):
            if self.match_search(i):
                self.index = i
                return
        self.index = -1

    # @override
    def process_input(self, key: int) -> int:
        # self.draw()

        if key == curses.KEY_UP:
            self.prev_option()
        elif key == curses.KEY_DOWN:
            self.next_option()
        elif key == curses.KEY_LEFT:
            self.search_text = self.search_text[:-1]
            self.update_index()
        elif 32 <= key < 127:
            self.search_text += chr(key).lower()
            self.update_index()
        elif key == curses.KEY_BACKSPACE:
            if len(self.search_text) > 0:
                self.search_text = self.search_text[:-1]
            else:
                self.search_text = []
            self.update_index()
        elif key == ord('\n'):
            if self.exit_on_action or self.options[self.index].exit_on_action:
                self.floating.enable = False
            if self.index != -1:
                self.options[self.index].action()
            return -1
        elif key == curses.KEY_EXIT:
            self.floating.enable = False
            if self.floating.exit_fn is not None:
                self.floating.exit_fn()
            return -1
        else:
            return key
        return -1


from tko.game.task_info import TaskInfo
from tko.logger.log_sort import LogSort
from tko.game.task_grader import TaskGrader
from typing import Any


class Resume:
    minutes_str: str = "minutes"
    versions_str: str = "versions"
    executions_str: str = "executions"
    lines_str: str = "lines"
    percent_str: str = "percent"

    def __init__(self, minutes: int = 0, versions: int = 0, executions: int = 0, lines: int = 0, percent: float = 0.0):
        self.minutes: int = minutes # elapsed time in minutes
        self.versions: int = versions # number of versions
        self.executions: int = executions # number of runs
        self.lines: int = lines # lines
        self.percent: float = percent # percent of completion

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        if self.minutes > 0:
            output[Resume.minutes_str] = str(self.minutes)
        if self.versions > 0:
            output[Resume.versions_str] = str(self.versions)
        if self.executions > 0:
            output[Resume.executions_str] = str(self.executions)
        if self.lines > 0:
            output[Resume.lines_str] = str(self.lines)
        if self.percent > 0.0:
            output[Resume.percent_str] = f"{self.percent:.2f}"
        return output
    
    def from_dict(self, info: dict[str, str]) -> None:
        # print(f"Loading Resume from dict: {info}")
        self.minutes    = int(info.get(Resume.minutes_str, 0))
        self.versions   = int(info.get(Resume.versions_str, 0))
        self.executions = int(info.get(Resume.executions_str, 0))
        self.lines      = int(info.get(Resume.lines_str, 0))
        percent_str     = info.get(Resume.percent_str, "0.0")
        if percent_str:
            self.percent = float(percent_str)


class TaskResume:
    def __init__(self, key: str):
        self.key: str = key
        self.resume: Resume = Resume()
        self.info: TaskInfo = TaskInfo()

    def from_log_sort(self, log_sort: LogSort):
        if log_sort.base_list:
            delta, _ = log_sort.base_list[-1]
            self.resume.minutes = round(delta.accumulated.total_seconds() / 60)
        self.resume.versions = len(log_sort.diff_list)
        self.resume.executions = len(log_sort.exec_list)
        if log_sort.diff_list:
            delta, last_diff = log_sort.diff_list[-1]
            self.resume.lines = last_diff.get_size()
        if log_sort.self_list:
            delta, last_self = log_sort.self_list[-1]
            self.info = last_self.get_info()
            self.resume.percent = TaskGrader(self.info).get_percent()
        return self
    
    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        output.update(self.resume.to_dict())
        output.update(self.info.get_kv())
        return output

    def from_dict(self, info: dict[str, str]) -> None:
        self.resume.from_dict(info)
        self.info.load_from_kv(info)



from typing import Any

class Log:
    _verbose = False
    
    @staticmethod
    def set_verbose(verbose: bool):
        Log._verbose = verbose

    @staticmethod
    def verbose(*args: str, **kwargs: Any):
        if Log._verbose:
            print(*args, **kwargs)

    @staticmethod
    def resume(*args: str, **kwargs: Any):
        if not Log._verbose:
            print(*args, **kwargs)
            print("", end="", flush=True)



from tko.settings.settings import Settings
from tko.settings.repository import Repository
from tko.play.task_graph import TaskGraph
from tko.util.raw_terminal import RawTerminal

class CmdTask:
    @staticmethod
    def show_graph(settings: Settings, rep: Repository, task_key: str, width: int | None = None, height: int | None = None):
        settings = settings
        rep = rep
        task_key = task_key
        if width is None:
            width = RawTerminal.get_terminal_size() // 2
        if height is None:
            height = round(width / 4)
        graph = TaskGraph(settings, rep, task_key, width, height).get_output()
        for line in graph:
            print(line)


from tko.enums.diff_mode import DiffMode
from tko.util.text import Text
from typing import Any
import curses

class AppSettings:

    def __init__(self):
        self.diff_mode = DiffMode.SIDE.value
        self.use_images = True
        self.use_borders = False
        self.editor = "code"
        self.timeout = 2
        self.last_tko_check_update = ""
        self.tko_last_version = ""
        self.key_left: int = curses.KEY_LEFT
        self.key_right: int = curses.KEY_RIGHT
        self.key_up: int = curses.KEY_UP
        self.key_down: int = curses.KEY_DOWN
        self.key_pg_up: int = curses.KEY_PPAGE
        self.key_pg_down: int = curses.KEY_NPAGE
        self.key_backspace: int = curses.KEY_BACKSPACE
        self.key_esc: int = 27

    def set_lastest_version(self, version: str):
        self.tko_last_version = version
        return self
    
    def get_lastest_version(self) -> str:
        return self.tko_last_version

    def set_latest_version_check_timestamp(self, date_str: str):
        self.last_tko_check_update = date_str
        return self
    
    def get_lastest_version_check_timestamp(self) -> str:
        return self.last_tko_check_update

    def set_key_left(self, key: int):
        self.key_left = key
        return self

    def set_key_right(self, key: int):
        self.key_right = key
        return self

    def set_key_up(self, key: int):
        self.key_up = key
        return self

    def set_key_down(self, key: int):
        self.key_down = key
        return self

    def get_key_left(self) -> int:
        return self.key_left

    def get_key_right(self) -> int:
        return self.key_right

    def get_key_up(self) -> int:
        return self.key_up

    def get_key_down(self) -> int:
        return self.key_down

    def set_key_pg_up(self, key: int):
        self.key_pg_up = key
        return self
    
    def set_key_pg_down(self, key: int):
        self.key_pg_down = key
        return self
    
    def set_key_esc(self, key: int):
        self.key_esc = key
        return self
    
    def set_key_backspace(self, key: int):
        self.key_backspace = key
        return self
    
    def get_key_esc(self) -> int:
        return self.key_esc
    
    def get_key_backspace(self) -> int:
        return self.key_backspace

    def get_key_pg_up(self) -> int:
        return self.key_pg_up

    def get_key_pg_down(self) -> int:
        return self.key_pg_down

    def to_dict(self):
        return self.__dict__

    def from_dict(self, attr_dict: dict[str, Any]):
        for key, value in attr_dict.items():
            if hasattr(self, key) and type(getattr(self, key)) == type(value):
                setattr(self, key, value)
        return self

    def toggle_diff(self):
        if self.diff_mode == DiffMode.SIDE.value:
            self.diff_mode = DiffMode.DOWN.value
        else:
            self.diff_mode = DiffMode.SIDE.value

    def toggle_borders(self):
        self.use_borders = not self.use_borders

    def toggle_images(self):
        self.use_images = not self.use_images


    def set_diff_mode(self, diff_mode: DiffMode):
        self.diff_mode = diff_mode.value
        return self

    def set_use_borders(self, borders: bool):
        self.use_borders = borders
        return self

    def set_use_images(self, images: bool):
        self.use_images = images
        return self

    def set_editor(self, editor: str):
        self.editor = editor
        return self

    def set_timeout(self, timeout: int):
        self.timeout = timeout
        return self

    def get_diff_mode(self) -> DiffMode:
        if self.diff_mode == DiffMode.SIDE.value:
            return DiffMode.SIDE
        return DiffMode.DOWN

    def get_use_images(self) -> bool:
        return self.use_images

    def get_use_borders(self) -> bool:
        return self.use_borders

    def get_editor(self) -> str:
        return self.editor

    def get_timeout(self) -> int:
        return self.timeout

    def __str__(self):
        output: list[str] = [str(Text.format("{g}", "Configurações globais:")),
                             "- Diff    : {}".format(str(self.get_diff_mode().value)),
                             "- Editor  : {}".format(self.get_editor()),
                             "- Bordas  : {}".format(self.get_use_borders()),
                             "- Images  : {}".format(self.get_use_images()),
                             "- Timeout : {}".format(self.get_timeout())]
        return "\n".join(output)


import os
from tko.util.decoder import Decoder

class FenoTitle:
    @staticmethod
    def extract_title(readme_file: str) -> str:

        folder = os.path.basename(os.path.dirname(readme_file))

        title = Decoder.load(readme_file).splitlines()[0]
        parts = title.split(" ")
        if parts[0].count("#") == len(parts[0]):
            del parts[0]
        title = " ".join(parts)
        return "@" + folder + ": " + title


class TaskInfo:
    class Keys:
        rate_str: str = "rate"
        ia_concept_str: str = "concept"
        ia_problem_str: str = "problem"
        ia_coding_str: str = "code"
        ia_debug_str: str = "debug"
        ia_refactor_str: str = "refactor"
        guided_str: str = "guided"

        study_str: str = "study" # study time spent in minutes
        friend_str: str = "friend" #
        feedback_str: str = "self" # feedback for the task

    def __init__(self):
        self.rate: int = 0

        self.study: int = 0  # quantos minutos ele estudou para fazer a atividade
        self.feedback: bool = False # se fez a auto avaliação
        self.friend: str = "" # se usou ajuda de monitor, amigo, etc

        self.guided: bool = False # se fez a atividade seguindo a aula presencial ou vídeo aula        
        self.ia_concept: bool = False # se usou IA para entender o conceito
        self.ia_problem: bool = False # se usou IA para resolver o problema
        self.ia_code: bool = False # se usou IA para escrever código
        self.ia_debug: bool = False # se usou IA para debugar
        self.ia_refactor: bool = False # se usou IA para refatorar

    def clone(self):
        return TaskInfo().load_from_kv(self.get_kv())

    def set_study(self, value: str):
        try:
            minutes = int(value)
            if minutes >= 0:
                self.study = minutes
        except ValueError:
            self.study = 0
        return self
    
    def set_rate(self, value: str):
        try:
            rate = int(value)
            if 0 <= rate <= 100:
                self.rate = rate
        except ValueError:
            self.rate = 0
        return self

 
    def load_from_kv(self, kv: dict[str, str]):
        if TaskInfo.Keys.rate_str in kv:
            self.set_rate(kv[TaskInfo.Keys.rate_str])
        if TaskInfo.Keys.study_str in kv:
            self.set_study(kv[TaskInfo.Keys.study_str])
        self.friend = kv.get(TaskInfo.Keys.friend_str, "")
        self.feedback = kv.get(TaskInfo.Keys.feedback_str, "0") == "1"
        self.guided = kv.get(TaskInfo.Keys.guided_str, "0") == "1"
        self.ia_concept = kv.get(TaskInfo.Keys.ia_concept_str, "0") == "1"
        self.ia_problem = kv.get(TaskInfo.Keys.ia_problem_str, "0") == "1"
        self.ia_code = kv.get(TaskInfo.Keys.ia_coding_str, "0") == "1"
        self.ia_debug = kv.get(TaskInfo.Keys.ia_debug_str, "0") == "1"
        self.ia_refactor = kv.get(TaskInfo.Keys.ia_refactor_str, "0") == "1"
        return self

    def get_kv(self) -> dict[str, str]:
        kv: dict[str, str] = {}
        if self.feedback:
            kv[TaskInfo.Keys.feedback_str] = "1"
        if self.rate != 0:
            kv[TaskInfo.Keys.rate_str] = str(self.rate)
        if self.study != 0:
            kv[TaskInfo.Keys.study_str] = str(self.study)
        if self.friend:
            kv[TaskInfo.Keys.friend_str] = self.friend
        if self.guided:
            kv[TaskInfo.Keys.guided_str] = "1"
        if self.ia_concept:
            kv[TaskInfo.Keys.ia_concept_str] = "1"
        if self.ia_problem:
            kv[TaskInfo.Keys.ia_problem_str] = "1"
        if self.ia_code:
            kv[TaskInfo.Keys.ia_coding_str] = "1"
        if self.ia_debug:
            kv[TaskInfo.Keys.ia_debug_str] = "1"
        if self.ia_refactor:
            kv[TaskInfo.Keys.ia_refactor_str] = "1"
        return kv
    
    


from tko.settings.repository import Repository
from tko.util.text import Text
from tko.util.to_asc import SearchAsc
from tko.play.flags import Flags
from tko.game.quest import Quest
from tko.game.task import Task
from tko.play.border import Border
from tko.down.drafts import Drafts
from tko.settings.settings import Settings
from tko.play.floating_manager import FloatingManager
from tko.util.symbols import symbols
from tko.game.tree_item import TreeItem

class TaskAction:
    BAIXAR   = "Baixar  "
    EXECUTAR = "Escolher"
    VISITAR  = "Visitar "
    EXPANDIR = "Expandir"
    CONTRAIR = "Contrair"
    BLOQUEIO = "Travado"

class TaskTree:

    def __init__(self, settings: Settings, rep: Repository, fman: FloatingManager):
        self.settings = settings
        self.app = settings.app
        self.game = rep.game
        self.rep = rep
        self.fman = fman
        self.style = Border(settings.app)
        self.colors = settings.colors
        self.items: list[TreeItem] = []
        self.all_items: dict[str, TreeItem] = {}
        self.selected_item: str = ""
        self.filler = "·"
        self.index_begin: int = 0
        self.search_text: str = ""
        self.expanded: set[str] = set()
        self.load_all_items()
        self.load_from_rep()
        self.update_tree(admin_mode = Flags.inbox.get_value() == Flags.inbox_all)
        self.MIN_TITLE_LENGTH = 50
        self.cache_max_title: None | int = None
        self.cache_task_times: dict[str, tuple[int, int]] = {}
        self.max_key_size: None | int = None
        self.reload_sentences()

    def get_max_key_size(self):
        keys_visible: list[str] = []
        if self.max_key_size is None:
            for q in self.game.quests.values():
                if not q.get_full_key() in self.expanded:
                    continue
                for t in q.get_tasks():
                    keys_visible.append(t.get_key())
        if len(keys_visible) == 0:
            return 10
        return max([len(k) for k in keys_visible])
    
    def get_max_quest_itens_key_size(self, quest_key: str):
        q = self.all_items.get(quest_key, None)
        if q is None or not isinstance(q, Quest):
            return 10
        keys: list[int] = [len(task.get_key()) for task in q.get_tasks()]
        if len(keys) == 0:
            return 10
        return max(keys)

    def load_all_items(self):
        for q in self.game.quests.values():
            self.all_items[q.get_full_key()] = q
            for t in q.get_tasks():
                self.all_items[t.get_full_key()] = t

    def load_from_rep(self):
        self.expanded = set(self.rep.data.expanded)
        self.selected_item = self.rep.data.selected

    def save_on_rep(self):
        self.rep.data.expanded = [x for x in self.expanded if x in self.game.quests.keys()]
        # self.rep.set_new_items([x for x in set(self.new_items)])
        self.rep.data.selected = self.selected_item
        tasks: dict[str, str] = {}
        for t in self.game.tasks.values():
            if len(t.info.get_kv()) != 0:
                tasks[t.get_full_key()] = t.save_to_db()
        self.rep.data.tasks = tasks

    def update_tree(self, admin_mode: bool):
        self.game.update_reachable_and_available()
        if not admin_mode:
            reachable_keys = [q.get_full_key() for q in self.game.quests.values() if q.is_reachable()]
            self.expanded = set(x for x in self.expanded if x in reachable_keys)

    def get_quest_title_size(self, quest: Quest) -> int:
        return len(quest.get_full_title()) + 6

    def get_task_title_size(self, task: Task) -> int:
        return len(task.get_full_title(self.get_max_key_size())) + 15

    def get_max_title(self) -> int:
        if self.cache_max_title:
            return self.cache_max_title
        
        items: list[int] = []
        for q in self.game.quests.values():
            items.append(self.get_quest_title_size(q))
            if q.get_full_key() in self.expanded:
                for t in q.get_tasks():
                    items.append(self.get_task_title_size(t))
        
        self.cache_max_title = self.MIN_TITLE_LENGTH
        if len(items) > 0:
            self.cache_max_title = max(items)
        self.cache_max_title = max(self.MIN_TITLE_LENGTH, self.cache_max_title)
        return self.cache_max_title

    def get_total_width(self) -> int:
        width = self.get_max_title()
        if Flags.show_time.is_true():
            width += len(self.format_hours_minutes("", 0, 0))
        width += 7 # for percent and symbols
        return max(width, self.MIN_TITLE_LENGTH + 10)

    @staticmethod
    def color_task_title(title: str, color: str) -> Text:
        words = title.split(" ")
        output = Text()
        for i, word in enumerate(words):
            if word.startswith("@") or word.startswith("#") or word.startswith("!"):
                output.addf(color + "g", word)
            elif word.startswith(":"):
                output.addf(color + "y", word)
            elif word.startswith("*"):
                output.addf(color + "c", word)
            elif word.startswith("+"):
                output.addf(color + "c", word)
            else:
                output.addf(color, word)
            if i < len(words) - 1:
                output.addf(color, " ")
        return output

    def format_hours_minutes(self, color: str, hours: int, minutes: int) -> Text:
        output = Text()
        if hours > 0 or minutes > 0:
            output.addf(color, f"{hours:02}h{minutes:02}m ")
        else:
            output.add("┄" * 6 + " ")
        return output

    def get_task_down_symbol(self, t: Task) -> Text.Token:
        if t.is_link():
            return symbols.task_view.set_fmt("c")

        if not t.is_import_type():
            return symbols.task_local
        
        down_symbol = symbols.task_to_download
        if self.is_downloaded_for_lang(t):
            down_symbol = symbols.task_downloaded
        return down_symbol
    
    def get_task_rate_symbol(self, t: Task) -> Text.Token:
        if t.task_rate == Task.TaskRate.AUTO:
            if t.info.feedback:
                return symbols.circle_filled.set_fmt("g")
            else:
                return symbols.circle_open
        if t.task_rate == Task.TaskRate.USER:
            if t.info.feedback:
                return symbols.task_human_filled.set_fmt("g")
            else:
                return symbols.task_human_open
        if t.task_rate == Task.TaskRate.TICK:
            if t.info.feedback:
                return symbols.square_filled.set_fmt("g")
            else:
                return symbols.square_open
        return Text.Token(" ")

    def get_task_path_symbol(self, t: Task) -> Text.Token:
        if t.task_path == Task.TaskPath.MAIN:
            return symbols.star_filled
        return symbols.star_open
    
    def get_task_rule_symbol(self, t: Task) -> Text.Token:
        if t.task_help == Task.TaskHelp.OPEN:
            return symbols.task_repeat.set_fmt("y")
        if t.task_help == Task.TaskHelp.EXAM:
            return symbols.task_denied.set_fmt("r")
        return Text.Token("")
    
    def get_task_action_symbol(self, t: Task) -> Text.Token:
        if t.task_action == Task.TaskAction.VIEW:
            return symbols.task_view.set_fmt("c")
        if t.task_action == Task.TaskAction.EDIT:
            return symbols.task_edit.set_fmt("c")
        return Text.Token(" ")

    def __str_task(self, focus_color: str, t: Task, lig_quest: str, quest_reachable: bool) -> Text:
        color_aval = "g" if quest_reachable and t.is_reachable() else "y"

        output = Text()
        output.addf("b", t.xp)
        output.addf(color_aval, lig_quest)
        
        output.add(self.get_task_down_symbol(t)).add(" ")
        rate = t.info.rate // 10
        output.add(t.get_prog_symbol(rate)).add(" ")
        output.add(self.get_task_rate_symbol(t)).add(" ")
        output.add(self.get_task_path_symbol(t)).add(" ")
        output.add(self.get_task_rule_symbol(t))

        in_focus = focus_color != ""
        # output.add(self.style.round_l(focus_color) if in_focus else " ")
        output.add(" ")
        color = "" if not in_focus else "k" + focus_color
        output.add(self.color_task_title(t.get_full_title(self.get_max_quest_itens_key_size(t.quest_key)), color))
        output.ljust(self.get_max_title(), Text.Token(" ", focus_color))
        # output.add(self.style.round_r(focus_color) if in_focus else " ") 

        if in_focus:
            output.add("◀ ")
        else:
            output.add("  ")

        if Flags.show_time.is_true():
            hours, minutes = self.get_task_hours_minutes(t)
            output.add(self.format_hours_minutes("g", hours, minutes))

        prog = round(t.get_percent())
        output.addf("y", str(prog).rjust(3, " ") + "%")
        return output

    def get_task_hours_minutes(self, task: Task) -> tuple[int, int]:
        if task.get_full_key() in self.cache_task_times:
            return self.cache_task_times[task.get_full_key()]
        logsort = self.rep.logger.tasks.task_dict.get(task.get_full_key(), None)
        if logsort is not None and len(logsort.base_list) > 0:
            delta, _ = logsort.base_list[-1]
            hours = delta.accumulated.seconds // 3600
            minutes = (delta.accumulated.seconds % 3600) // 60
            self.cache_task_times[task.get_full_key()] = (hours, minutes)
            return hours, minutes
        self.cache_task_times[task.get_full_key()] = (0, 0)
        return 0, 0
    
    def __str_quest(self, has_kids: bool, focus_color: str, q: Quest) -> Text:
        con = "┄┄"
        n_visible, n_hidden = self.count_visible_hidden_tasks(q)
        if n_visible > 0:
            con = "━─" if n_hidden == 0 else "━┄"
            if q.get_full_key() in self.expanded and has_kids:
                con = "─╮" if n_hidden == 0 else "┄╮"

        color_reachable = "g" if q.is_reachable() else "y"
        # if Flags.quests.get_value() == Flags.quest_enable:
        #     for quest in self.game.quests.values():
        #         if not quest.is_reachable() and q.get_db_key() in quest.requires:
        #                 color_reachable = "y"
        #                 break

        output: Text = Text().addf(color_reachable, con)

        in_focus = focus_color != ""

        if focus_color == "":
            item_key = self.selected_item
            if item_key in list(self.game.quests.keys()):
                quest = self.game.quests[item_key]
                if not quest.is_reachable():
                    if q.get_full_key() in quest.requires:
                        focus_color = "y"

        # if in_focus:
            # output.add(self.style.round_l(focus_color))
        output.add(" ")
        title = q.get_full_title()
        if focus_color:
            for i, _ in enumerate(title.data):
                title.data[i].set_fmt(focus_color)
    
        output.add(title)
        output = output.ljust(self.get_max_title(), Text.Token(self.filler, focus_color))
        # output.add(self.style.round_r(focus_color) if in_focus else " ")
        
        if in_focus:
            output.add("◀ ")
        else:
            output.add("  ")
        
        if Flags.show_time.is_true():
            hours, minutes = self.get_quest_time(q)
            output.add(self.format_hours_minutes("g", hours, minutes))
        output.add(q.get_resume_by_percent())

        return output

    def get_quest_time(self, quest: Quest) -> tuple[int, int]:
        hours = 0
        minutes = 0
        for t in quest.get_tasks():
            th, tm = self.get_task_hours_minutes(t)
            hours += th
            minutes += tm
        hours += minutes // 60
        minutes = minutes % 60
        return hours, minutes

    
    def __get_focus_color_quest(self, item: Quest) -> str:
        if not item.is_reachable():
                return "R"
        return self.settings.colors.focused_item

    def filter_by_search(self) -> tuple[set[str], str | None]:
        matches: set[str] = set()
        search = SearchAsc(self.search_text)
        first: None | str = None
        
        for quest in self.game.quests.values():
            if search.inside(quest.get_title()):
                first = first or quest.get_full_key()
                matches.add(quest.get_full_key())
            for task in quest.get_tasks():
                if search.inside(task.get_title()):
                    first = first or task.get_full_key()
                    matches.add(quest.get_full_key())
                    matches.add(task.get_full_key())
        return matches, first

    def __try_add(self, filtered: set[str], matcher: SearchAsc, item: TreeItem):
        if self.search_text == "":
            self.items.append(item)
            return True
        if item.get_full_key() in filtered:
            pos = matcher.find(item.get_sentence().get_str())
            found = pos != -1
            if found:
                for i in range(pos, pos + len(self.search_text)):
                    item.get_sentence().data[i].fmt = "Y"
            self.items.append(item)
            return True
        return False

    def is_hide_tasks(self, task: Task) -> bool:
        if Flags.inbox.get_value() == Flags.inbox_only and task.task_path == Task.TaskPath.SIDE and not self.is_downloaded_for_lang(task):
            return True
        return False

    def remove_empty_quests(self):
        new_expanded = set(self.expanded)
        for q in self.game.quests.values():
            if q.get_full_key() not in self.expanded:
                continue
            visible, _ = self.count_visible_hidden_tasks(q)
            if visible == 0:
                new_expanded.remove(q.get_full_key())
        self.expanded = new_expanded

    def reload_sentences(self):
        self.cache_task_times.clear()
        self.cache_max_title = None # force recalculation of max title size
        self.items = []
        self.remove_empty_quests()
        filtered, _ = self.filter_by_search()
        matcher = SearchAsc(self.search_text)

        hide: bool = Flags.inbox.get_value() == Flags.inbox_only

        quests = [q for q in self.game.quests.values() if q.get_full_key() in self.game.quests.keys() if q.get_full_key() in filtered]
        if hide:
            quests = [q for q in quests if q.is_reachable()]

        for q in quests:
            tasks =[t for t in q.get_tasks() if t.get_full_key() in filtered]
            if hide:
                tasks = [t for t in tasks if t.is_reachable()]
            quest_focus_color: str = self.__get_focus_color_quest(q)
            focus_color = quest_focus_color if self.selected_item == q.get_full_key() else ""
            q.set_sentence(self.__str_quest(len(tasks) > 0, focus_color, q))

            full_task_size = len(tasks)
            tasks = [t for t in tasks if not self.is_hide_tasks(t)]
            filtered_tasks_size = len(tasks)
            has_hidden_tasks = full_task_size != filtered_tasks_size

            self.__try_add(filtered, matcher, q)
            if q.get_full_key() in self.expanded:
                for t in tasks:
                    if has_hidden_tasks:
                        ligq = "┆ " if t != tasks[-1] else "╰ "
                    else:
                        ligq = "├ " if t != tasks[-1] else "╰ "
                    focus_color: str = quest_focus_color if self.selected_item == t.get_full_key() else ""
                    t.set_sentence(self.__str_task(focus_color, t, ligq, q.is_reachable()))
                    # if self.is_hide_tasks(t):
                    #     continue
                    self.__try_add(filtered, matcher, t)
        # verifying if it has any selected item
        if self.items:
            found = False
            for item in self.items:
                if item.get_full_key() == self.selected_item:
                    found = True
                    break
            if not found:
                self.selected_item = self.items[0].get_full_key()
                self.reload_sentences()

    def process_collapse_all(self):
        if any([q in self.expanded for q in self.game.quests.keys()]):
            self.expanded = set([key for key in self.expanded if key not in self.game.quests.keys()])
        else:
            self.expanded = set()

    def process_expand_all(self):
        # if any cluster outside expanded
        new_expanded = [x for x in self.expanded]
        for qkey in self.game.quests.keys():
            if qkey not in self.expanded:
                new_expanded.append(qkey)
        self.expanded = set(new_expanded)

    def  get_selected_index(self) -> int:
        for i, item in enumerate(self.items):
            if item.get_full_key() == self.selected_item:
                return i
        return 0

    def get_selected_throw(self) -> TreeItem:
        if len(self.items) > 0:
            index = self.get_selected_index()
            return self.items[index]
        raise IndexError("No item selected")
            

    def set_selected_by_index(self, index: int):
        if index < 0:
            index = 0
        if index >= len(self.items):
            index = len(self.items) - 1
        self.selected_item = self.items[index].get_full_key()

    def arrow_right(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        obj = self.items[index]

        if isinstance(obj, Quest):
            if not self.unfold(obj):
                while True:
                    index += 1
                    if index >= len(self.items):
                        break
                    obj = self.items[index]
                    if isinstance(obj, Quest):
                        break
                self.set_selected_by_index(index)
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest):
                    break
                if index == len(self.items) - 1:
                    break
                index += 1
            self.set_selected_by_index(index)

    def walk_left_on_quest(self, index: int):
        while True:
            if index == 0:
                break
            index -= 1
            obj = self.items[index]
            if isinstance(obj, Quest) and obj.get_full_key() in self.expanded:
                break
            if index == 0:
                break
        self.set_selected_by_index(index)

    def walk_left_on_cluster(self, index: int):
        while True:
            if index == 0:
                break
            index -= 1
            obj = self.items[index]
            if isinstance(obj, Quest):
                break
            if index == 0:
                break
        self.set_selected_by_index(index)

    def arrow_left(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        obj = self.items[index]
        if isinstance(obj, Quest):
            if not self.fold(obj):
                self.walk_left_on_quest(index)
            return
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest):
                    break
                index -= 1
            self.set_selected_by_index(index)

    def is_admin_mode(self) -> bool:
        return Flags.inbox.get_value() == Flags.inbox_all

    def unfold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest) and not obj.is_reachable() and not self.is_admin_mode():
            return False
        if isinstance(obj, Quest):
            visible, _ = self.count_visible_hidden_tasks(obj)
            if visible == 0:
                return False
        if isinstance(obj, Quest):
            if obj.get_full_key() not in self.expanded:
                self.expanded.add(obj.get_full_key())
                return True
        return False

    def count_visible_hidden_tasks(self, quest: Quest) -> tuple[int, int]:
        visible = 0
        hidden = 0
        for t in quest.get_tasks():
            if not self.is_hide_tasks(t):
                visible += 1
            else:
                hidden += 1
        return visible, hidden

    def fold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest):
            if obj.get_full_key() in self.expanded:
                self.expanded.remove(obj.get_full_key())
                return True
        return False

    def toggle(self, obj: Quest):
            if not self.fold(obj):
                self.unfold(obj)

    def get_senteces(self, dy: int) -> list[Text]:
        index = self.get_selected_index()
        if len(self.items) < dy:
            self.index_begin = 0
        else:
            if index < self.index_begin:  # subiu na tela
                self.index_begin = index
            elif index >= dy + self.index_begin:  # desceu na tela
                self.index_begin = index - dy + 1

        sentences: list[Text] = []
        for i in range(self.index_begin, len(self.items)):
            sentences.append(self.items[i].get_sentence())
        return sentences

    def move_up(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        self.set_selected_by_index(index - 1)
        

    def move_down(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        self.set_selected_by_index(min(len(self.items) - 1, index + 1))

    # return a TaskAction
    def get_task_action(self, task: Task) -> tuple[str, str]:
        if task.is_link():
            return "B", TaskAction.VISITAR
        if task.is_static_type():
            return "G", TaskAction.EXECUTAR
        if not self.is_downloaded_for_lang(task):
            return "Y", TaskAction.BAIXAR
        return "G", TaskAction.EXECUTAR
        
    def is_downloaded_for_lang(self, task: Task):
        folder = task.get_workspace_folder()
        if folder is None:
            return False
        
        lang = self.rep.data.lang
        drafts = Drafts.load_drafts_only(folder, lang)
        if drafts:
            return True
        return False
        


from tko.play.frame import Frame
from tko.util.text import Text
from tko.play.fmt import Fmt
from typing import Callable
import enum
import curses
from abc import ABC, abstractmethod


class FloatingType(enum.Enum):
    WARNING = "warning"
    ERROR = "error"


class FloatingABC(ABC):
    @abstractmethod
    def draw(self):
        pass

    @abstractmethod
    def process_input(self, key: int) -> int:
        pass

    @abstractmethod
    def is_enable(self) -> bool:
        pass

class Floating(FloatingABC):
    def __init__(self):
        self.frame = Frame(0, 0)
        self.content: list[Text] = []
        self.type: FloatingType = FloatingType.WARNING
        self.enable = True
        self.exit_fn: Callable[[], None] | None = None
        self.exit_key: None | int = None
        self.centralize_text = True
        self.floating_align = ""

    def left(self):
        self.floating_align = [c for c in self.floating_align if c != ">"]
        self.floating_align += "<"
        return self
    
    def right(self):
        self.floating_align = [c for c in self.floating_align if c != "<"]
        self.floating_align += ">"
        return self
    
    def top(self):
        self.floating_align = [c for c in self.floating_align if c != "v"]
        self.floating_align += "^"
        return self
    
    def bottom(self):
        self.floating_align = [c for c in self.floating_align if c != "^"]
        self.floating_align += "v"
        return self

    def disable(self):
        self.enable = False

    def set_text_ljust(self):
        self.centralize_text = False
        return self

    def set_text_center(self):
        self.centralize_text = True
        return self

    def set_frame_align(self, _align: str):
        self.floating_align = _align
        return self

    def set_header(self, text: str):
        self.frame.set_header(Text().addf("/", text), "")
        return self
    
    def set_header_text(self, sentence: Text):
        self.frame.set_header(sentence, "")
        return self
    
    def set_footer(self, text: str):
        self.frame.set_footer(Text().addf("/", text), "")
        return self
    
    def set_footer_text(self, sentence: Text):
        self.frame.set_footer(sentence, "")
        return self
    
    def set_exit_key(self, key: str):
        self.exit_key = ord(key)
        return self

    def set_exit_fn(self, fn: Callable[[], None]):
        self.exit_fn = fn
        return self

    def _set_xy(self, dy: int, dx: int):
        valid = "<>^v"
        for c in self.floating_align:
            if c not in valid:
                raise ValueError("Invalid align " + c)

        lines, cols = Fmt.get_size()

        x = (cols - dx) // 2
        if "<" in self.floating_align:
            x = 1
        elif ">" in self.floating_align:
            x = cols - dx - 2

        y = (lines - dy) // 2
        if "^" in self.floating_align:
            y = 1
        elif "v" in self.floating_align:
            y = lines - dy - 2

        self.frame.set_pos(y, x)
        return self
            
    def is_enable(self):
        return self.enable

    def calc_dy_dx(self):
        header_len = self.frame.get_header().len()
        footer_len = self.frame.get_footer().len()
        data = [x.len() for x in self.content] + [header_len, footer_len]
        max_dx = max(data)
        dx = max_dx
        dy = len(self.content)
        return dy, dx

    def setup_frame(self):
        dy, dx = self.calc_dy_dx()
        self.frame.set_inner(dy, dx)
        self._set_xy(dy, dx)
        self.frame.set_fill()

    def put_text(self, text: str):
        lines = text.splitlines()
        for line in lines:
            self.content.append(Text().add(line))
        return self

    def put_sentence(self, sentence: Text):
        for line in sentence.split('\n'):
            self.content.append(line)
        return self
    
    def set_content(self, content: list[str]):
        self.content = [Text().add(x) for x in content]
        return self

    def set_default_footer(self):
        if self.frame.get_footer().len() == 0:
            label = Text().addf("/", " Pressione espaço ")
            self.frame.set_footer(label, "", "─", "─")
        return self

    def set_default_header(self):
        if self.frame.get_header().len() == 0:
            if self.type == FloatingType.WARNING:
                self.set_header(" Aviso ")
            elif self.type == FloatingType.ERROR:
                self.set_header(" Erro ")

    def set_warning(self):
        self.type = FloatingType.WARNING
        self.frame.set_border_color("y")
        return self


    def set_error(self):
        self.type = FloatingType.ERROR
        self.frame.set_border_color("r")
        return self

    def draw(self):
        self.set_default_header()
        self.set_default_footer()
        self.setup_frame()
        self.frame.draw()
        self.write_content()

    def write_content(self):
        y = 0
        for line in self.content:
            x = 0
            if self.centralize_text:
                x = (self.frame.get_dx() - line.len()) // 2
            self.frame.write(y, x, line)
            y += 1
        return self

    def process_input(self, key: int) -> int:
        if self.type == FloatingType.WARNING or self.type == FloatingType.ERROR:
            if key == curses.KEY_RESIZE:
                return -1
            self.enable = False
            if self.exit_fn is not None:
                self.exit_fn()
            if self.exit_key is not None:
                return self.exit_key
            if key == ord(" ") or key == curses.KEY_EXIT: # evita propagar a tecla
                return -1
            return key
        return -1



from tko.game.quest import Quest
from tko.game.task import Task
from tko.game.game_builder import GameBuilder
from tko.game.game_validator import GameValidator
from tko.settings.rep_source import RepSource
# from typing import override
from icecream import ic # type: ignore
import re

def load_html_tags(task: str) -> None | str:
    pattern = r"<!--\s*(.*?)\s*-->"
    match = re.search(pattern, task)
    if not match:
        return None
    return match.group(1).strip()


class Game:
    def __init__(self):
        self.sources: list[RepSource] = []
        self.ordered_quests: list[str] = [] # ordered clusters
        self.quests: dict[str, Quest] = {}  # quests indexed by quest key
        self.tasks: dict[str, Task] = {}  # tasks indexed by task key

    def get_task(self, key: str) -> Task:
        if key in self.tasks:
            return self.tasks[key]
        raise Warning(f"fail: tarefa '{key}' não encontrada no curso")

    def get_xp_resume(self):
        total = 0
        obtained = 0
        for q in self.quests.values():
            o, t = q.get_xp()
            total += t
            obtained += o
        return obtained, total

    def get_skills_resume(self) -> tuple[dict[str, float], dict[str, float], dict[str, float]]:
        # obtained, priority, complete
        priority: dict[str, float] = {}
        complete: dict[str, float] = {}
        obtained: dict[str, float] = {}

        for q in self.quests.values():
            for t in q.get_tasks():
                for skill, value in t.skills.items():
                    gvalue = (value * t.get_xp() * t.get_ratio())
                    if gvalue < 0.1:
                        gvalue = 0
                    obtained[skill] = obtained.get(skill, 0) + gvalue
                    if not t.is_optional():
                        priority[skill] = priority.get(skill, 0) + value * t.get_xp()
                    complete[skill] = complete.get(skill, 0) + value * t.get_xp()

        # remove all keys with value 0
        # obtained = {k: v for k, v in obtained.items() if v > 0}
        # priority = {k: v for k, v in priority.items() if v > 0}
        # complete = {k: v for k, v in complete.items() if v > 0}
        return obtained, priority, complete

    def sum_xp(self, obtained: dict[str, float], priority: dict[str, float], complete: dict[str, float]) -> tuple[float, float, float]:
        total_obtained = 0
        total_priority = 0
        total_complete = 0
        for key, value in complete.items():
            total_obtained += obtained.get(key, 0)
            total_priority += priority.get(key, 0)
            total_complete += value
        return total_obtained, total_priority, total_complete

    def get_sandbox_source(self) -> RepSource:
        for s in self.sources:
            if s.is_sandbox_source():
                return s
        raise ValueError("Local sandbox source not found")

    def set_sources(self, sources: list[RepSource], language: str, silent: bool = False) -> None:
        self.sources = sources
        self.ordered_clusters = []
        self.clusters = {}
        self.quests = {}
        self.tasks = {}
        self.language = language
        self.silent = silent
        self.load_sources()
    
    def load_sources(self):
        for source in self.sources:
            gb = GameBuilder(source)
            gb.build_from(self.language)
            for quest_key in gb.ordered_quests:
                self.ordered_quests.append(source.name + "@" + quest_key)

            gb_quests = gb.collect_quests()
            gb_tasks = gb.collect_tasks()
            for quest in gb_quests.values():
                self.quests[quest.get_full_key()] = quest
            for task in gb_tasks.values():
                self.tasks[task.get_full_key()] = task
        GameValidator(self.quests).validate()

        # for t in self.tasks.values():
        #     t.get_link(os.path.dirname(filename) + "/")

    @staticmethod
    def is_reachable_quest(q: Quest, cache: dict[str, bool]):
        if q.get_full_key() in cache:
            return cache[q.get_full_key()]

        if len(q.requires_ptr) == 0:
            cache[q.get_full_key()] = True
            return True
        cache[q.get_full_key()] = all([r.is_complete() and Game.is_reachable_quest(r, cache) for r in q.requires_ptr])
        return cache[q.get_full_key()]

    # def __get_reachable_quests(self):
    #     # cache needs to be reseted before each call
    #     cache: dict[str, bool] = {}
    #     return [q for q in self.quests.values() if Game.__is_reachable_quest(q, cache)]

    def update_reachable_and_available(self):
        for q in self.quests.values():
            q.set_reachable(False)
            q.update_tasks_reachable()
        for q in self.quests.values():
            q.set_reachable(False)

        cache: dict[str, bool] = {}
        for q in self.quests.values():
            if Game.is_reachable_quest(q, cache):
                q.set_reachable(True)
                q.set_reachable(True)

    # @override
    def __str__(self):
        output: list[str] = []
        for q in self.quests.values():
            output.append("  - " + str(q))
            for t in q.get_tasks():
                output.append("    - " + str(t))
        output.append(100 * "-")
        for q in self.quests.values():
            output.append(str(q))
        output.append(100 * "-")
        for t in self.tasks.values():
            output.append(str(t))
        return "\n".join(output)


from __future__ import annotations
from pathlib import Path
from typing import Any
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.logger.logger import Logger
from tko.logger.task_resume import TaskResume
from tko.logger.log_sort import LogSort
from tko.play.daily_graph import DailyGraph
from tko.mico.collected import Collected, Game
from tko.util.text import Text
from tko.settings.rep_paths import RepPaths

import yaml # type: ignore
import json
import os
import argparse
import csv
    

class CmdCollect:
    @staticmethod
    def resume(rep: Repository) -> dict[str, TaskResume]:
        logger: Logger = rep.logger
        tasks: dict[str, LogSort] = logger.tasks.task_dict
        resume_dict: dict[str, TaskResume] = {}

        for key in tasks:
            log_sort = tasks[key]
            resume = TaskResume(key).from_log_sort(log_sort)
            resume_dict[key] = resume

        return resume_dict

    @staticmethod
    def daily_graph(rep: Repository, width: int, height: int, colored: bool) -> str:
        dg = DailyGraph(rep.logger, width, height)
        image = dg.get_graph()
        if not colored:
            return "\n".join([x.get_str() for x in image])
        return "\n".join([str(x) for x in image])

    @staticmethod
    def load_game_as_quest_list(rep: Repository) -> list[Game.Quest]:
        game = rep.game
        if not game:
            return []
        output: list[Game.Quest] = []

        for quest in game.quests.values():
            output_quest = Game.Quest(quest.get_full_key())
            output.append(output_quest)
            for task in quest.get_tasks():
                output_quest.tasks.append(Game.Task(key=task.get_full_key(), value=task.xp, is_leet=task.is_leet(), opt=task.is_optional()))
        return output
    
    @staticmethod
    def collect_batch(args: argparse.Namespace):
        git_repo_list: list[Path] = [Path(x) for x in args.path]
        CollectMany.execute(git_repo_list, json_path=args.json, csv_path=args.csv, block_prefix=args.block_prefix)

    @staticmethod
    def collect_main(args: argparse.Namespace):
        params = CollectSingle.CollectParams()
        params.folder = Path() if args.changedir is None else Path(args.changedir)
        params.width = args.width
        params.height = args.height
        params.daily = args.daily
        params.resume = args.resume
        params.game = args.game
        params.log = args.log
        params.json_output = args.json
        params.colored = args.color
        data: Collected = CollectSingle.collect(params)

        if params.json_output:
            # print(yaml.dump(data))
            print(json.dumps(data.to_dict(), indent=4, ensure_ascii=False))


    @staticmethod
    def update(args: argparse.Namespace):
        folder = args.folder
        if not os.path.isdir(folder):
            print(f"Folder {folder} does not exist.")
            return
        rep = Repository(folder)
        if not rep.found():
            print(f"Folder {folder} is not a valid tko repository.")
            return
        rep.load_config().load_game()
        print(f"Repositório cache atualizado.")

    @staticmethod
    def list(_args: argparse.Namespace):
        settings = Settings()
        print(f"SettingsFile\n- {settings.settings_file}")
        print(str(settings))


    @staticmethod
    def reset(args: argparse.Namespace):
        _ = args
        sp = Settings().reset()
        print(sp.settings_file)
        sp.save_settings()

    # @staticmethod
    # def graph(args):
    #     rep = Repository(args.folder).load_config().load_game()
    #     rep.game.check_cycle()
    #     Graph(rep.game).generate()

class CollectSingle:
    class CollectParams:
        def __init__(self):
            self.folder: Path = Path()
            self.width: int = 10
            self.height: int = 10
            self.daily: bool = False
            self.resume: bool = False
            self.game: bool = False
            self.log: bool = False
            self.json_output: bool = False
            self.colored: int = 1




    @staticmethod
    def collect_to_json(repo_folder: Path, daily: bool = True, resume: bool = True, game: bool = True) -> str:
        params = CollectSingle.CollectParams()
        params.folder = repo_folder
        params.daily = daily
        params.resume = resume
        params.game = game
        params.json_output = True # dont echo
        results: Collected = CollectSingle.collect(params)
        return json.dumps(results.to_dict(), indent=4, ensure_ascii=False)

    @staticmethod
    def collect(param: CollectParams) -> Collected:
        rep = Repository(param.folder)
        if not rep.found():
            path = os.path.abspath(param.folder)
            print(f"Repository not found in {path}")
            return Collected()
        rep.set_global_cache().load_config().load_game(try_update=False, silent=True)
        data = Collected()

        if param.daily:
            graph = CmdCollect.daily_graph(rep, param.width, param.height, param.colored == 1)
            data.graph = graph
            if not param.json_output:
                print(graph)

        if param.resume:
            resume_data = CmdCollect.resume(rep)
            data.resume = resume_data
            if not param.json_output:
                for key, value in resume_data.items():
                    print(f"{key}: {value.to_dict()}")

        if param.log:
            log_data = sorted(rep.logger.history.get_entries().items(), key=lambda x: x[0])
            data.log = [str(entry) for _, entry in log_data]
            if not param.json_output:
                for _, entry in log_data:
                    print(entry)

        if param.game:
            game_data = CmdCollect.load_game_as_quest_list(rep)
            data.quests = game_data
            if not param.json_output:
                for quest in game_data:
                    print(str(quest))
        return data

class CollectMany:
    @staticmethod
    def find_common_prefix(folders: list[str]) -> str:
        if not folders:
            return ""
        common = ""
        for chars in zip(*folders):
            if all(c == chars[0] for c in chars):
                common += chars[0]
            else:
                break
        return common


    @staticmethod
    def execute(git_dir_list: list[Path], json_path: str | None = None, csv_path: str | None = None, block_prefix: str | None = None):
        git_dir_list = [git_dir for git_dir in git_dir_list if git_dir.is_dir()]
        common_prefix = CollectMany.find_common_prefix([str(folder.name) for folder in git_dir_list])

        usernames = [repo.name[len(common_prefix):].strip("/\\") for repo in git_dir_list]
        padding = max(len(username) for username in usernames) + 1

        output_map: dict[str, Any] = {}
        for git_dir, username in zip(git_dir_list, usernames):
            tko_rep_folder_list = RepPaths.rec_search_for_repo_subdir(git_dir)
            if not tko_rep_folder_list:
                print(Text("").addf("r", f"{username: <{padding}}").addf("r", f"TKO repo not found in {git_dir}"))
                continue
            tko_folder = tko_rep_folder_list[0]
            multiple_found = Text().addf("r", " - Multiple TKO repos found, using the first one." if len(tko_rep_folder_list) > 1 else "")
            print(Text("").addf("g" if not multiple_found else "y", f"{username: <{padding}}").addf("", f"Running tko collect in {tko_folder}").add(multiple_found))
            output = CollectSingle.collect_to_json(tko_folder, daily=False, resume=True, game=False)

            try:
                json_output: dict[str, Any] = json.loads(output) if output != "" else {}
            except json.JSONDecodeError:
                print(Text("").addf("r", f"{username: <{padding}}").addf("r", "Error: Failed to parse JSON output"))
                continue
            if "error" in json_output:
                print(Text("").addf("r", f"{username: <{padding}}").addf("r", f"Error: {json_output['error']}"))
                continue

            output_map[username] = json_output["resume"] if "resume" in json_output else {}
        if json_path is not None:
            with open(json_path, "w", encoding="utf-8") as f:
                print(Text("").addf("g", f"Saving extracted data to {json_path}"))
                json.dump(output_map, f, indent=4, ensure_ascii=False)
        
        header_keys = ["username", "key", "minutes", "versions", "executions", "rate", "study", "self", "friend", "concept", "problem", "code", "debug", "refactor", "guided"]
        if block_prefix is not None:
            header_keys = ["block"] + header_keys
        if csv_path is not None:
            with open(csv_path, "w", encoding="utf-8", newline="") as f:
                writer = csv.DictWriter(f, fieldnames=header_keys)
                writer.writeheader()
                for student_key, info in output_map.items():
                    for key, data in info.items():
                        if "@" in key:
                            key = key.split("@")[1]

                        row: dict[str, Any] = {
                            "username": student_key,
                            "key": key,
                            "minutes": data.get("minutes", 0),
                            "versions": data.get("versions", 0),
                            "executions": data.get("executions", 0),
                            "rate": round(float(data.get("rate", 0.0))),
                            "study": data.get("study", 0),
                            "self": data.get("self", ""),
                            "friend": data.get("friend", ""),
                            "concept": data.get("concept", ""),
                            "problem": data.get("problem", ""),
                            "code": data.get("code", ""),
                            "debug": data.get("debug", ""),
                            "refactor": data.get("refactor", ""),
                            "guided": data.get("guided", "")
                        }

                        if block_prefix is not None:
                            row["block"]= f"{block_prefix}"
                        writer.writerow(row)
            print(Text("").addf("g", f"Saving extracted data to {csv_path}"))


import os
from pathlib import Path
from platformdirs import user_cache_dir

class RepPaths:
    CFG_FILE = "repository.yaml"
    OLD_HISTORY_FILE = "history.csv"
    TASK_LOG_FILE = "task_log.csv"
    DAILY_FILE = "daily.yaml"
    TRACK_FOLDER = "track"
    LOG_FOLDER = "log"
    CACHE_FOLDER = "cache"
    CONFIG_FOLDER = ".tko"
    use_global_cache_folder: bool =  False # teacher tools, to avoid rewrite cache in student repos

    def __init__(self, repo_dir: Path | str):
        self.root_folder: Path = Path(repo_dir) if isinstance(repo_dir, str) else repo_dir

    @staticmethod
    def walk_to_root(start: Path):
        current: Path = start.resolve()
        while True:
            yield current
            if current.parent == current:
                break
            current = current.parent

    @staticmethod
    def rec_search_for_repo_subdir(folder: Path) -> list[Path]:
        repos: list[Path] = [
            cfg.parent.parent
            for cfg in folder.rglob(f"{RepPaths.CONFIG_FOLDER}/{RepPaths.CFG_FILE}")
        ]
        return repos

    @staticmethod
    def rec_search_for_repo_parents(folder: Path) -> Path | None:
        for path in RepPaths.walk_to_root(folder):
            target: Path = path / RepPaths.CONFIG_FOLDER / RepPaths.CFG_FILE
            if target.is_file():
                return path

        return None
    
    def get_track_folder(self) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.TRACK_FOLDER

    def get_log_folder(self) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.LOG_FOLDER

    def get_cache_folder(self) -> Path:
        if RepPaths.use_global_cache_folder:
            return Path(user_cache_dir("tko")) / RepPaths.CACHE_FOLDER
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.CACHE_FOLDER

    def get_track_task_folder(self, label: str) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.TRACK_FOLDER / label

    def get_config_folder(self) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER

    def get_config_file(self) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.CFG_FILE

    def get_config_backup_file(self) -> Path:
        return Path(str(self.get_config_file()) + ".backup")

    def get_old_history_file(self) -> Path:
        return self.root_folder / RepPaths.CONFIG_FOLDER / RepPaths.OLD_HISTORY_FILE

    def get_workspace_dir(self) -> Path:
        return self.root_folder

    def has_local_config_file(self) -> bool:
        return os.path.exists(self.get_config_file())


# # from datetime import datetime
# from unittest.mock import MagicMock
# from tko.util.logger import Logger, LoggerMemory, ActionData.Type, ActionData

# class TestLogger:

#     def setUp(self):
#         self.logger_store = LoggerMemory()
#         self.logger = Logger(self.logger_store)
#         self.logger.set_history_file("test_rep")


#     def test_record_event(self):
#         self.logger.record_other_event(ActionData.Type.OPEN, "task1", "payload1")
#         entries = self.logger_store.get_action_entries()
#         self.assertEqual(len(entries), 1)
#         self.assertEqual(entries[0].action_value, ActionData.Type.OPEN.value)
#         self.assertEqual(entries[0].task_key, "task1")
#         self.assertEqual(entries[0].payload, "payload1")

#     def test_get_last_hash(self):
#         self.logger.record_other_event(ActionData.Type.OPEN, "task1", "payload1")
#         last_hash = self.logger.get_last_hash()
#         self.assertEqual(last_hash, self.logger_store.entries[0].hash)

#     def test_check_log_file_integrity(self):
#         self.logger.record_other_event(ActionData.Type.OPEN, "task1", "payload1")
#         self.logger.record_other_event(ActionData.Type.QUIT, "task2", "payload2")
#         integrity_issues = self.logger.check_log_file_integrity()
#         self.assertEqual(len(integrity_issues), 0)

#     def test_store_in_cached(self):
#         action_data = ActionData(datetime.now().strftime('%Y-%m-%d %H:%M:%S'), ActionData.Type.SELF.value, "task1", "payload1")
#         self.assertTrue(self.logger.store_in_cached(action_data))
#         self.logger.cached_action = action_data
#         self.assertTrue(self.logger.store_in_cached(action_data))

#     def test_record_action_data(self):
#         action_data = ActionData("um", ActionData.Type.SELF.value, "task1", "payload1")
#         self.logger.record_action_data(action_data)
#         action_data = ActionData("dois", ActionData.Type.SELF.value, "task1", "payload2")
#         self.logger.record_action_data(action_data)
#         self.assertEqual(len(self.logger_store.entries), 0)
#         action_data = ActionData("tres", ActionData.Type.SELF.value, "task2", "payload1")
#         self.logger.record_action_data(action_data)
#         self.assertEqual(len(self.logger_store.entries), 1)
#         action_data = ActionData("quatro", ActionData.Type.QUIT.value, "test_rep")
#         self.logger.record_action_data(action_data)
#         self.assertEqual(len(self.logger_store.entries), 3)
#         self.assertEqual(self.logger_store.entries[0].task_key, "task1")
#         self.assertEqual(self.logger_store.entries[1].task_key, "task2")
#         self.assertEqual(self.logger_store.entries[2].action_value, ActionData.Type.QUIT.value)

#     def test_check_log_file_integrity2(self):
#         self.logger.record_other_event(ActionData.Type.OPEN, "task1", "payload1")
#         self.logger.record_other_event(ActionData.Type.QUIT, "task2", "payload2")
#         self.logger.record_other_event(ActionData.Type.QUIT, "task3", "payload3")
#         integrity_issues = self.logger.check_log_file_integrity()
#         self.assertEqual(len(integrity_issues), 0)
#         self.logger_store.entries[1].hash = "wrong_hash"
#         integrity_issues = self.logger.check_log_file_integrity()
#         self.assertEqual(len(integrity_issues), 1)

# if __name__ == '__main__':
#     unittest.main()


from tko.play.flags import Flag
from tko.play.floating_manager import FloatingManager
from tko.play.floating import Floating
from typing import Callable


class FlagFunctor:
    def __init__(self, flag: Flag, fman: FloatingManager):
        self.flag = flag
        self.fman = fman

    def __call__(self):
        self.flag.toggle()
        index = self.flag.get_index()
        if index < len(self.flag.get_msgs()):
            msg = self.flag.get_msgs()[index]
            self.fman.add_input(Floating().set_warning().put_text(msg))

class GradeFunctor:
    def __init__(self, grade: int, fn: Callable[[int], None]):
        self.grade: int = grade
        self.fn = fn

    def __call__(self):
        self.fn(self.grade)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from subprocess import run, PIPE
import os
import configparser
import argparse
import tempfile

class Const:
    ansi_green = "\033[92m"
    ansi_reset = "\033[0m"

class Problem:
    tag_value = "value"
    tag_param = "param"
    tag_partial = "partial"


class MyTest:
    def __init__(self, label: str, value: int = 1, partial: bool = True, param: str = "--track"):
        self.label: str = label
        self.awarded: float = 0      # Percentage score (0-100)
        self.value: float = value    # Used to calculate the weighted mean
        self.param: str = param         # Additional parameters for the test
        self.partial: bool = partial # Whether the test can be partially scored

    def set_percentage(self, text: str):
        if text[-1] == "%":
            value = float(text[:-1])
            if self.partial:
                self.awarded = value
            else:
                self.awarded = value if value == 100 else 0
        return self

    
    def run(self):
        _, temp_file = tempfile.mkstemp(prefix="temp_result_file", suffix=".txt")
        extra: list[str] = []
        if self.param:
            extra = self.param.split(" ")
        running_prefix()
        result = run(["tko", "eval", "--timeout", "30", "--none"] + extra + ['-r', temp_file, f"src/{self.label}"], stderr=PIPE, text=True)
        if result.returncode != 0:
           if result.stderr != "":
               print(result.stderr)
        if result.returncode == 0:
            if os.path.isfile(temp_file):
                percent = open(temp_file, "r").read().splitlines()[0].strip()
                self.set_percentage(percent)
                os.remove(temp_file)
        return self


class Grading:
    def __init__(self):
        self.tests: list[MyTest] = []
    def add_and_run_test(self, test: MyTest):
        self.tests.append(test)
        test.run()

    def calc_grade(self) -> int:
        total_weight = sum(test.value for test in self.tests)
        max_label_len = max(len(test.label) for test in self.tests) + 1
        grade: float = 0
        grading_prefix()
        print(f"{'TestCases':<{max_label_len}}| passed | value | earned")
        sep = f"{'-' * max_label_len}|--------|-------|-------"
        grading_prefix()
        print(sep)
        for test in self.tests:
            test.value = test.value * 100 / total_weight
            awarded = test.awarded * test.value / 100
            grading_prefix()
            print(f"{test.label.ljust(max_label_len)}|   {round(test.awarded):3d}% |  {round(test.value):3d}% |   {round(awarded):3d}%")
            grade += awarded
        grading_prefix()
        print(sep)
        grading_prefix()
        print(f"{'Total':<{max_label_len}}|        |  100% |    {Const.ansi_green}{round(grade):3d}%{Const.ansi_reset}")
        return round(grade)

    @staticmethod
    def load_config(config_file: str) -> list[MyTest]:
        config = configparser.ConfigParser()
        config.read(config_file)
        problems: list[MyTest] = []
        for section in config.sections():
            problem = MyTest(
                label=section,
                value=config.getint(section, Problem.tag_value, fallback=1),
                param=config.get(section, Problem.tag_param, fallback=""),
                partial=config.getboolean(section, Problem.tag_partial, fallback=True)
            )
            problems.append(problem)
        return problems
    
    @staticmethod
    def load_readme(readme_file: str) -> list[MyTest]:
        problems: list[MyTest] = []
        if not os.path.isfile(readme_file):
            print(f"README file '{readme_file}' does not exist.")
            return problems
        
        with open(readme_file, "r") as f:
            lines = f.readlines()
        
        for line in lines:
            if not line.startswith("- [ ] `"):
                continue
            test = MyTest("")
            parts = line.split("`")
            title = parts[1].strip()
            words = title.split(" ")
            for w in words:
                if w.startswith("@"):
                    test.label = w[1:]
                elif w.startswith("*"):
                    test.value = int(w[1:])
            if test.label != "":
                problems.append(test)
        return problems

    @staticmethod
    def main(args: argparse.Namespace):
        problems: list[MyTest] = []
        if args.config is not None:
            problems = Grading.load_config(args.config)

        if args.readme is not None:
            problems = Grading.load_readme(args.readme)


        if len(problems) == 0:
            print("No problems found in the configuration file.")
            return

        test_list = Grading()
        for problem in problems:
            test_list.add_and_run_test(problem)

        awarded = test_list.calc_grade()
        if args.output is not None:
            with open(args.output, "w") as f:
                f.write(str(awarded))

def running_prefix():
    print("[TKO RUNNING] ", flush=True, end='')

def grading_prefix():
    print("[TKO GRADING] ", flush=True, end='')



from __future__ import annotations
from pathlib import Path
from typing import Any
import os
from enum import Enum

from tko.settings.git_cache import GitCache

class SourceType(Enum):
    LOCAL_FILE = "local"
    GIT_SOURCE = "git"

class RepSource:
    STUDENT_SANDBOX_NAME = "sandbox"

    DEFAULT_DRAFT_FOLDER = "00_rascunhos"

    AUTOLOAD_CLUSTER_COMMAND = "autoload_cluster"
    AUTOLOAD_QUEST_COMMAND = "autoload_quest"

    sandbox_readme_content = f"""
# Sandbox

Cada vez que criar um rascunho, ele será salvo na pasta `{DEFAULT_DRAFT_FOLDER}` dentro do sandbox.

Você pode criar quantos rascunhos quiser. Eles serão rastreados pela chave key no preâmbulo do rascunho.

Renomeie o nome da pasta para renomear o rascunho. Exclua a pasta para excluir o rascunho.

Sinta-se à vontade para organizar seus rascunhos em subpastas dentro do sandbox, se desejar.

## {STUDENT_SANDBOX_NAME}
<!--{AUTOLOAD_CLUSTER_COMMAND}=.-->
"""[1:]
    

    class Keys:
        NAME = "name"
        TARGET = "target"
        TYPE = "type"
        WRITEABLE = "writeable"
        QUESTS = "quests"
        TASKS = "tasks"
        BRANCH = "branch"

    """
    alias: str - unique identifier for the source, used to reference the source in the code and configuration
    target: str - for LOCAL_FILE, it's folder path; for GIT_CLONE, it's the git repository link
    source_type: SourceType - type of the source, either LOCAL_FILE or GIT_CLONE
    writeable: bool - indicates if the source is writeable, used to determine if the source can be modified or not
    branch: str | None - for GIT_CLONE type, it's the branch to clone or pull, default is "master"
    filters: list[str] | None - list of filters to apply when loading quests from the source, used to filter quests based on certain criteria
    """
    def __init__(self, alias: str):
        self.name = alias
        self.__target = ""
        self.source_type: SourceType = SourceType.LOCAL_FILE
        self.__writeable: bool = False
        self.branch: str | None = None
        self.quests: list[str] | None = None
        self.tasks: list[str] | None = None
        self.rep_local_workspace: Path | None = None   # if read-only rep, rep folder to tasks will be, join(local_workspace, alias)
        self.rep_cache_folder: Path | None = None 

    def set_local_source(self, target: Path, writeable: bool = False):
        self.source_type = SourceType.LOCAL_FILE
        self.__target = str(target)
        self.__writeable = writeable
        return self
    
    def is_sandbox_source(self) -> bool:
        return self.name == RepSource.STUDENT_SANDBOX_NAME # for backward compatibility, to remove in the future
    
    def set_student_sandbox(self):
        self.name = RepSource.STUDENT_SANDBOX_NAME
        self.set_local_source(target=Path(RepSource.STUDENT_SANDBOX_NAME), writeable=True)
        return self
    
    def get_default_sandbox_draft_folder(self) -> str:
        return self.DEFAULT_DRAFT_FOLDER
    
    def ensure_sandbox_source(self, local_workspace: Path):
        # create folder sandbox inside rep_workspace if not exists
        sandbox_path = local_workspace / RepSource.STUDENT_SANDBOX_NAME
        sandbox_path.mkdir(parents=True, exist_ok=True)
        readme_file = sandbox_path / "README.md"
        if not os.path.exists(readme_file):
            with open(readme_file, "w") as f:
                f.write(self.sandbox_readme_content)
        return self
    
    # return quests and tasks filters
    def get_filters(self) -> tuple[list[str] | None, list[str] | None]:
        return self.quests, self.tasks

    def set_git_source(self, target: str, branch: str | None = None):
        self.source_type = SourceType.GIT_SOURCE
        self.__target = target
        self.branch = branch
        return self
    
    def is_local(self) -> bool:
        return self.source_type == SourceType.LOCAL_FILE

    def set_writeable(self, writeable: bool = True):
        self.__writeable = writeable
        return self

    def get_writeable(self) -> bool:
        return self.__writeable
    
    def is_read_only(self) -> bool:
        if self.source_type == SourceType.LOCAL_FILE:
            return not self.__writeable
        if self.source_type == SourceType.GIT_SOURCE:
            return True
        return True

    def get_url_link(self) -> str:
        if self.source_type == SourceType.GIT_SOURCE:
            return self.__target
        elif self.source_type == SourceType.LOCAL_FILE:
            return self.__target
        raise ValueError("Unknown source type")

    def is_git_source(self) -> bool:
        return self.source_type in [SourceType.GIT_SOURCE, "clone"] # for backward compatibility, to remove in the future

    def is_local_source(self) -> bool:
        return self.source_type == SourceType.LOCAL_FILE

    def get_source_readme(self) -> Path:
        if self.source_type == SourceType.LOCAL_FILE:
            
            file: str = os.path.join(self.__target, "README.md")
            if os.path.isabs(file):
                return Path(file)
            else:
                return self.get_rep_workspace() / file
        if self.source_type == SourceType.GIT_SOURCE:
            return self.get_source_folder() / "README.md"
        raise ValueError("Unknown source type")
    
    def get_source_folder(self) -> Path:
        if self.source_type == SourceType.LOCAL_FILE:
            return Path(self.__target)
        if self.source_type == SourceType.GIT_SOURCE:
            git_cache = GitCache(self.get_rep_cache_folder())
            return git_cache.get(self.get_url_link(), force_update=False)# absolute path to cached repo
        raise ValueError("Unknown source type")

    
    def set_filters(self, quests: list[str] | None, tasks: list[str] | None = None):
        self.quests = quests
        self.tasks = tasks
        return self
    
    def set_rep_globals(self, local_workspace: Path, cache_folder: Path):
        self.rep_local_workspace = local_workspace
        self.rep_cache_folder = cache_folder

    def get_rep_cache_folder(self) -> Path:
        if self.rep_cache_folder is None:
            raise ValueError("Local cache folder is not set")
        return self.rep_cache_folder

    def get_rep_workspace(self) -> Path:
        if self.rep_local_workspace is None:
            raise ValueError("Local workspace is not set")
        return self.rep_local_workspace
    
    def get_source_workspace(self) -> Path:
        return self.get_rep_workspace() / self.name
    
    def get_task_workspace(self, task_key: str) -> Path:
        if not self.is_read_only():
            raise ValueError("Source is not read-only, task workspace is the same as source workspace")
        return self.get_source_workspace() / task_key

    def load_from_dict(self, data: dict[str, Any]):
        Keys = RepSource.Keys
        if Keys.NAME in data and isinstance(data[Keys.NAME], str):
            self.name = data[Keys.NAME]
        if "alias" in data and isinstance(data["alias"], str): # for backward compatibility
            self.name = data["alias"]
        if "database" in data and isinstance(data["database"], str): # for backward compatibility
            self.name = data["database"]
        if Keys.TARGET in data and isinstance(data[Keys.TARGET], str):
            self.__target = data[Keys.TARGET]
        if "link" in data and isinstance(data["link"], str): # for backward compatibility
            self.__target = data["link"]
            if self.__target.endswith("README.md"):
                self.__target = os.path.dirname(self.__target)
        if Keys.BRANCH in data and isinstance(data[Keys.BRANCH], str):
            self.branch = data[Keys.BRANCH]
        else:
            self.branch = "master"
        if Keys.TYPE in data and isinstance(data[Keys.TYPE], str):
            type_str = data[Keys.TYPE]
            if type_str == SourceType.LOCAL_FILE.value:
                self.source_type = SourceType.LOCAL_FILE
            else:
                self.source_type = SourceType.GIT_SOURCE
        else:
            self.source_type = SourceType.LOCAL_FILE
        if Keys.QUESTS in data and isinstance(data[Keys.QUESTS], list):
            self.quests = data[Keys.QUESTS]
        if "filters" in data and isinstance(data["filters"], list): # for backward compatibility
            self.quests = data["filters"]
        if Keys.TASKS in data and isinstance(data[Keys.TASKS], list):
            self.tasks = data[Keys.TASKS]
        if Keys.WRITEABLE in data and isinstance(data[Keys.WRITEABLE], bool):
            self.__writeable = data[Keys.WRITEABLE]
        if self.name == RepSource.STUDENT_SANDBOX_NAME or self.name == "local": # for backward compatibility, to remove in the future
            self.__writeable = True
        return self

    def save_to_dict(self) -> dict[str, Any]:
        Keys = RepSource.Keys
        output: dict[str, Any] = {
            Keys.NAME: self.name,
            Keys.TARGET: self.__target,
            Keys.TYPE: self.source_type.value,
            Keys.WRITEABLE: self.__writeable,
        }
        if self.branch is not None and self.branch != "master":
            output[Keys.BRANCH] = self.branch
        output[Keys.QUESTS] = self.quests
        output[Keys.TASKS] = self.tasks
        return output
